from sympy import *

r=Symbol('r')
el=Symbol('el')
az=Symbol('az')
a_n = Symbol('a_n')
a_e = Symbol('a_e')
a_z = Symbol('a_z')
c = Symbol('c')

def norm(arr):
  return sqrt(arr[0]*arr[0] + arr[1]*arr[1] + arr[2]*arr[2] )
  
S = [r,el,az]
dt_0 = r/c
ant_nez = [a_n,a_e,a_z]
S_nez = [r*cos(az)*cos(el), r*sin(az)*cos(el), r*sin(el)]
delta_nez = [S_nez[0] - ant_nez[0], S_nez[1] - ant_nez[1], S_nez[2] - ant_nez[2]]
dt_ant = norm(delta_nez) / c

dt = dt_ant - dt_0

print(simplify(dt))

from math import sin,cos
def delay(r, el, az, a_n, a_e, a_z):
  c = 2.99793e8
  return (-r + (a_e**2 - 2.0*a_e*r*sin(az)*cos(el) + a_n**2 - 2.0*a_n*r*cos(az)*cos(el) + a_z**2 - 2.0*a_z*r*sin(el) + r**2)**(1.0/2.0))/c




def assert_equal(x,y):
  dx = abs(x-y)
  if  dx > 1e-5*(abs(x) + abs(y)):
    print("FAIL %g != %g (%f)" % (x,y,dx))
  else:
    print("PASS %g == %g (%f)" % (x,y,dx))
   
r = 1e10
def to_rad(xx):
  return xx*3.1415926/180.0

c = 2.99793e8
r = 1e9

assert_equal(delay(r, to_rad(90.0), to_rad(45), 0.0, 1.0, 0.0), 0.0)
assert_equal(delay(r, to_rad(45.0), to_rad(0), 0.0, 1.0, 0.0), 0.0)
assert_equal(delay(r, to_rad(0.0), to_rad(90), 0.0, 1.0, 0.0), -1.0/c)
assert_equal(delay(r, to_rad(45.0), to_rad(45), 0.0, 1.0, 0.0), -1.6678e-9)



import math
import datetime
from .gps_time import GpsTime
from tart.util import utc 

def get_utc_hours(timestamp):
    sec =    (timestamp - datetime.datetime.combine(timestamp.date(), datetime.time.min)).total_seconds()
    hours = (sec / 3600.0)
    return hours

def sign(x):
    return math.copysign(1, x)

def get_julian_date(timestamp):
    if (timestamp.tzinfo is not None):
        if not isinstance(timestamp.tzinfo, utc.UTC):
            raise RuntimeError('Timestamp has timezone {}'.format(timestamp.tzinfo))
    
    YY=timestamp.year
    MM=timestamp.month
    DD=timestamp.day
    UTcor=0
    HR=timestamp.hour
    Min=timestamp.minute
    Sec=timestamp.second
    return 367*YY - (7*(YY+((MM+9)//12))//4) + (275*MM//9)+ DD + 1721013.5 + UTcor//24 \
        - 0.5*sign((100*YY)+MM-190002.5) + 0.5 + HR/24.0 + Min/(60.0*24.0) + Sec/(3600.0*24.0)

def get_mjd(timestamp):
    return get_julian_date(timestamp)-2400000.5


def JulianDay (utc_date):
    return get_julian_date(utc_date)



from tart.operation import observation
from tart.imaging import visibility
from tart.util import angle

import numpy as np
from scipy.signal import hilbert

from tart.imaging.correlator import *

TEST_SCOPE_CONFIG = 'tart/test/test_telescope_config.json'

import unittest
class TestCorrelator(unittest.TestCase):
  def test_correlator_simp(self):
    from tart.operation import settings
    from tart.simulation import simulation_source
    from tart.util import angle
    from tart.util import constants
    from tart.simulation import radio
    import datetime


    import numpy as np
    N = 20
    a0 = np.random.randint(0,2,2**N)
    a1 = np.random.randint(0,2,2**N)
    self.assertAlmostEqual(corr_b(a0,a1,N), corr_b_pat(a0,a1))
    
    from tart.operation import observation
    from tart.operation import settings
    import datetime
    t = datetime.datetime.utcnow()
    c = settings.from_file(TEST_SCOPE_CONFIG)
    c.Dict['num_antenna'] = 2
    d = [a0,a1]
    o = observation.Observation(timestamp=t, config=c, data=d)
    
    cor = Correlator(van_vleck_corr=True)
    vis_a = cor.correlate(o,mode='roll')
    #vis_b = cor.correlate_roll(o)
    vis_c = cor.correlate(o,mode='fftw_hilbert')
    #print(vis_a.vis(0,1), vis_b.vis(0,1), vis_c.vis(0,1))

    sample_duration = 16.02e-1
    sample_duration = 4e-1
    config = settings.from_file(TEST_SCOPE_CONFIG)
    rad = radio.Max2769B(noise_level = np.ones(config.get_num_antenna()))
    src = simulation_source.SimulationSource(amplitude = 1., azimuth = angle.from_dms(0.), elevation = angle.from_dms(5.), sample_duration = sample_duration)
    fc0 = rad.int_freq

    int_sig = np.exp(-2.0j*np.pi*fc0*rad.baseband_timebase)

    def add_noise(sig):
      return sig + np.random.normal(0.,0.1,len(sig))

    def sig2binary(signal):
      return np.array([1. if (x>=0) else 0. for x in signal])

    antsig1 = src.s_baseband(rad.baseband_timebase) * int_sig

    for fraction in np.random.uniform(-np.pi/3,np.pi/3,10):
      dt = ((2.*np.pi)/src.omega)*fraction
      print('dt', dt)
      antsig2 = (src.s_baseband(rad.baseband_timebase+dt) * int_sig * np.exp(1.0j*src.omega*dt))

      antsig1 = add_noise(antsig1.real)
      antsig2 = add_noise(antsig2.real)

      d = [sig2binary(antsig1),sig2binary(antsig2)]
      obs = observation.Observation(timestamp=t, config=c, data=d)

      vis_a = cor.correlate(o,mode='roll')
      #vis_b = cor.correlate_roll(o)
      vis_c = cor.correlate(o,mode='fftw_hilbert')
      vis_d = cor.correlate(o,mode='fftw_hilbert_sign')
      
      #print(vis_a.vis(0,1),vis_b.vis(0,1))
      print(vis_a.vis(0,1),vis_c.vis(0,1))
      print(vis_a.vis(0,1),vis_d.vis(0,1))
      #print(vis_b.vis(0,1),vis_c.vis(0,1))
      #print(vis_b.vis(0,1),vis_d.vis(0,1))
      print(vis_c.vis(0,1),vis_d.vis(0,1))
      
      #self.assertAlmostEqual(vis_a.vis(0,1),vis_b.vis(0,1),4)
      self.assertAlmostEqual(vis_a.vis(0,1),vis_c.vis(0,1),4)
      self.assertAlmostEqual(vis_a.vis(0,1),vis_d.vis(0,1),4)
      #self.assertAlmostEqual(vis_b.vis(0,1),vis_c.vis(0,1),4)
      #self.assertAlmostEqual(vis_b.vis(0,1),vis_d.vis(0,1),4)
      self.assertAlmostEqual(vis_c.vis(0,1),vis_d.vis(0,1),4)
             

      
      vis = cor.correlate(obs,mode='roll')
      cor_out = angle.from_dms(angle.wrap_360(np.angle(vis.v[0], deg=True)))
      input_angle = angle.from_dms(angle.wrap_360(dt*src.omega*180./np.pi)) #.to_degrees()
      #print(cor_out, input_angle)
      #print(type(cor_out),  type(input_angle))
      #print(cor_out - input_angle)
      self.assertLess(np.abs((cor_out - input_angle).to_degrees()), 30.)

      #d = [sig2binary(antsig1),sig2binary(antsig2)]
      #obs = observation.Observation(timestamp=t, config=c, data=d)
      #vis = cor.correlate(obs,mode='fftw_hilbert')
      #cor_out = angle.wrap_360(np.angle(vis.v[0], deg=True))
      #print(cor_out, input_angle)
      #self.assertTrue(np.abs(cor_out - input_angle) < 20.)

class TestHilbert(unittest.TestCase):
  def test_hilbert(self):
    import scipy.signal
    print('hilbert')
    x = np.sin(np.arange(10000))
    h_scipy = scipy.signal.hilbert(x)
    h_pyfftw = hilbert(x)
    [self.assertAlmostEqual(i,j, 6) for i, j in zip(h_scipy,h_pyfftw)]


if __name__ == '__main__':
    unittest.main()




import unittest

import numpy as np
import datetime
import matplotlib.pyplot as plt
from tart.util import angle

from tart.simulation.simulation_source import *

class TestSimulationSource(unittest.TestCase):
  def setUp(self):
      self.crab = SimulationSource(r=1e9, amplitude = 1.0, azimuth = angle.from_dms(0.0), elevation = angle.from_dms(45.0), sample_duration = 1e-3) # 0.1 ms sample duration
      self.s2 = SimulationSource(r=1e9, amplitude = 1.0, azimuth = angle.from_dms(0.0), elevation = angle.from_dms(90.0), sample_duration = 1e-3) # 0.1 ms sample duration
      self.s3 = SimulationSource(r=1e9, amplitude = 1.0, azimuth = angle.from_dms(0.0), elevation = angle.from_dms(0.0), sample_duration = 1e-3) # 0.1 ms sample duration

  def test_full_simp(self):
    tend= 1e-4
    full_time = np.arange(0,tend,1./(1.575e9))
    crab_s = self.crab.s(full_time)

    samp_time = np.arange(0,tend,1./(16.e6))
    samp_time2 = np.arange(0,tend,1./(16.e6))+1e-6

    crab_s_baseband = self.crab.s_baseband(samp_time)
    crab_s_baseband2 = self.crab.s_baseband(samp_time2)
    plt.plot(full_time,crab_s/np.cos(full_time * self.crab.omega), label='full')
    plt.plot(samp_time,crab_s_baseband, label='sampled')
    plt.plot(samp_time2,crab_s_baseband2, label='sampled')


    plt.legend()
    #plt.show()

  def test_signal_deterministic(self):
    samp_freq = 16.e6
    samp_time = np.arange(0,1e-4,1./(samp_freq))
    a = self.crab.s_baseband(samp_time)
    dt = 1./(16.e6)
    b = self.crab.s_baseband(samp_time+dt)

    for i in np.abs(a[1:]-b[:-1]):
      self.assertLess(i,1e-5)

    dt = 0.5/(16.e6)
    b = self.crab.s_baseband(samp_time+dt)
    for i in range(1,100):
      print(i)
      if a[i+1] - a[i] > 0:
        self.assertTrue(a[i+1] - a[i] > a[i+1] - b[i])
        self.assertTrue(a[i+1] - a[i] > b[i] - a[i])

  def test_signal(self):
    timebase = np.arange(0,1e-3,2e-10)
    crab_signal = self.crab.s(timebase)
    mean = crab_signal.mean()
    stdev = crab_signal.std()
    self.assertLess(abs(mean), 1.0e-6)
    self.assertLess(stdev, 0.40)  # Should be 0.33
    self.assertGreater(stdev, 0.31)

  def test_baseband(self):
    timebase = np.arange(0,1e-3,1e-7)
    crab_signal = self.crab.s_baseband(timebase).real
    mean = crab_signal.mean()
    stdev = crab_signal.std()
    self.assertLess(abs(mean), 2.0e-2)
    self.assertLess(stdev, 0.55) # Should be 0.5
    self.assertGreater(stdev, 0.47)




import pickle

from tart.imaging import uvfitsgenerator
from tart.imaging import radio_source
from tart.imaging import location

from tart.simulation import antennas
from tart.util import skyloc
from tart.util import constants
from tart.util import angle

import numpy as np
#import pyfftw.interfaces.numpy_fft as fft
import numpy.fft as fft
import time

import os
import copy

import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable

cc = np.concatenate


def get_max_ang(nw, num_bin):
    ret = np.degrees(num_bin/(4.*nw))
    return ret

class Synthesis_Imaging(object):
    def __init__(self, cal_vis_list, fixed_zenith=True):
        self.cal_vis_list = cal_vis_list
        self.fixed_zenith = fixed_zenith
        if self.fixed_zenith:
            self.phase_center = None
        else:
            vt = self.cal_vis_list[0]
            ra, dec = location.get_loc(vt.get_config()).horizontal_to_equatorial(vt.get_timestamp(), angle.from_dms(90), angle.from_dms(0))
            self.phase_center = radio_source.CosmicSource(ra, dec, 1e10)
        self.grid_file = 'grid.idx'
        self.grid_idx = None
        #print('debug:' , self.phase_center.to_horizontal(vt.config.get_loc(),vt.timestamp))

    def set_grid_file(self,fpath):
        self.grid_file = fpath

    def get_uvfits(self):
        gen = uvfitsgenerator.UVFitsGenerator(copy.deepcopy(self.cal_vis_list), self.phase_center)
        return gen


    def get_difmap_movie(self, base_index, frames):
        os.system("rm out.uvfits")
        fits_name = "out.uvfits" #self.fname + ".uvfits"
        for i_frame in frames:
            v_list = copy.deepcopy(self.cal_vis_list)
            vis_indexes = base_index + i_frame
            vis_part = [v_list[ni] for ni in vis_indexes]

            # ra, dec = vis_part[0].config.get_loc().horizontal_to_equatorial(vis_part[0].timestamp, angle.from_dms(90.), angle.from_dms(0.))
            # self.phase_center = radio_source.CosmicSource(ra, dec)
            # print(ra, dec)

            uvgen = uvfitsgenerator.UVFitsGenerator(vis_part, self.phase_center) # FIXME
            uvgen.write(fits_name)
            difcmd = get_difmap(fits_name, i_frame)
            f = open('difmap_cmds', 'w')
            f.write(difcmd)
            f.close()
            os.system("difmap < difmap_cmds")
            os.system("rm out.uvfits")

    def get_uuvvwwvis_zenith(self):
        vis_l = []
        #for cal_vis in copy.deepcopy(self.cal_vis_list[:1]):
        for cal_vis in self.cal_vis_list[:1]:
            ant_p = np.array(cal_vis.get_config().get_antenna_positions())
            bls = cal_vis.get_baselines()
            pos_pairs = ant_p[np.array(bls)]
            uu_a, vv_a, ww_a = (pos_pairs[:,0] - pos_pairs[:,1]).T/constants.L1_WAVELENGTH
            for bl in bls:
                ant_i, ant_j = bl
                vis_l.append(cal_vis.get_visibility(ant_i,ant_j))
        return uu_a, vv_a, ww_a, np.array(vis_l)

    def get_grid_idxs(self,uu_a, vv_a, num_bin, nw):
        try:
            if self.grid_idx is None:
                self.grid_idx = pickle.load(open(self.grid_file, 'rb'))
                #print('finished loading ' + self.grid_file)
        except:
            print('generating...')
            uu_edges = np.linspace(-nw, nw, num_bin+1)
            vv_edges = np.linspace(-nw, nw, num_bin+1)
            grid_idx = []
            for uu, vv in zip(uu_a, vv_a):
                i = uu_edges.__lt__(uu).sum()-1
                j = vv_edges.__lt__(vv).sum()-1
                i2 = uu_edges.__lt__(-uu).sum()-1
                j2 = vv_edges.__lt__(-vv).sum()-1
                grid_idx.append([i,j,i2,j2])
            self.grid_idx = np.array(grid_idx)
            save_ptr = open(self.grid_file, 'wb')
            pickle.dump(self.grid_idx, save_ptr, pickle.HIGHEST_PROTOCOL)
            save_ptr.close()
        return self.grid_idx


    def get_uvplane_zenith(self, num_bin = 1600, nw = 36,):
        uu_a, vv_a, ww_a, vis_l = self.get_uuvvwwvis_zenith()
        arr = np.zeros((num_bin, num_bin), dtype=np.complex64)
        # place complex visibilities in the UV grid and prepare averaging by counting entries.
        grid_idxs = self.get_grid_idxs(uu_a, vv_a, num_bin, nw)
        count_arr = np.zeros((num_bin, num_bin), dtype=np.int16)
        #for k, v_l in enumerate(vis_l):
        #    i,j,i2,j2 = grid_idxs[k]
        #    count_arr[j, i] += 1
        #    count_arr[j2, i2] += 1
        if count_arr.max()>1:
            # apply the masked array and divide by number of entries
            for k, v_l in enumerate(vis_l):
                i,j,i2,j2 = grid_idxs[k]
                arr[j, i] += v_l
                arr[j2, i2] += np.conjugate(v_l)
            n_arr = n_arr/(count_arr)
        else:
            arr[grid_idxs[:,1],grid_idxs[:,0]] = vis_l
            arr[grid_idxs[:,3],grid_idxs[:,2]] = np.conjugate(vis_l)
        n_arr = np.ma.masked_array(arr[:, :], count_arr.__lt__(1.))
        return n_arr

    def get_uvplane(self, num_bin = 1600, nw = 36, grid_kernel_r_pixels=0.5, use_kernel=False):

        for cal_vis in (self.cal_vis_list):
            if self.fixed_zenith:
                uu_a, vv_a, ww_a = cal_vis.get_all_uvw()
                vis_l, bls = cal_vis.get_all_visibility()
            else:
                uu_l = []
                vv_l = []
                ww_l = []
                vis_l = []

                ts = cal_vis.get_timestamp()
                ra, dec = self.phase_center.radec(ts)
                c = cal_vis.get_config()
                ant_p = np.asarray(c.get_antenna_positions())
                loc = location.get_loc(c)
                bls = cal_vis.get_baselines()

                for bl in bls:
                    ant_i, ant_j = bl
                    a0 = antennas.Antenna(loc, ant_p[ant_i])
                    a1 = antennas.Antenna(loc, ant_p[ant_j])
                    uu, vv, ww = antennas.get_UVW(a0, a1, ts, ra, dec)
                    uu_l.append(uu/constants.L1_WAVELENGTH)
                    vv_l.append(vv/constants.L1_WAVELENGTH)
                    ww_l.append(ww/constants.L1_WAVELENGTH)
                    vis_l.append(cal_vis.get_visibility(ant_i,ant_j))
                uu_a = np.asarray(uu_l)
                vv_a = np.asarray(vv_l)
                ww_a = np.asarray(ww_l)
                vis_l = np.asarray(vis_l)

        uu_edges = np.linspace(-nw, nw, num_bin+1)
        vv_edges = np.linspace(-nw, nw, num_bin+1)

        if use_kernel==False:

            n_arr = np.zeros((num_bin, num_bin), dtype=np.complex64)
            uu_comb = np.concatenate((uu_a, -uu_a))
            vv_comb = np.concatenate((vv_a, -vv_a))
            all_v     = np.concatenate((vis_l, np.conjugate(vis_l)))
            h_real,_,_ = np.histogram2d(vv_comb, uu_comb, weights = all_v.real, bins=[vv_edges, uu_edges])
            h_imag,_,_ = np.histogram2d(vv_comb, uu_comb, weights = all_v.imag, bins=[vv_edges, uu_edges])
            num_entries,_,_ = np.histogram2d(vv_comb, uu_comb, bins=[vv_edges, uu_edges])
            n_arr[:,:] = (h_real+(1j*h_imag))
            pos = np.where(num_entries.__gt__(1))
            n_arr[pos] /= num_entries[pos]

        else:
            pixels_per_wavelength = num_bin/(nw*2.)
            halfbin = float(nw)/(num_bin)
            mid_points_uv = np.mgrid[-nw+halfbin:nw-halfbin:num_bin*1j, -nw+halfbin:nw-halfbin:num_bin*1j]
            n_arr = np.zeros((num_bin, num_bin), dtype=complex)

            r_noise_wavelengths = 0.1
            grid_kernel_r_wavelength = r_noise_wavelengths + grid_kernel_r_pixels / pixels_per_wavelength
            offset_px = np.ceil(grid_kernel_r_wavelength * pixels_per_wavelength)
            offsets = np.arange(-offset_px, offset_px+1)


            vis_max_abs = np.max(np.abs(vis_l))
            for uu, vv, v_l in zip(uu_a, vv_a, vis_l):
                i = uu_edges.__lt__(uu).sum()-1
                j = vv_edges.__lt__(vv).sum()-1
                # print('u', mid_points_uv[0][i-1,0], mid_points_uv[0][i,0], mid_points_uv[0][i+1,0], uu)
                # print('v', mid_points_uv[1][0,j], vv)
                for i_offset in offsets:
                    for j_offset in offsets:
                        r = np.sqrt(np.power(uu-mid_points_uv[0][i+i_offset,0],2) + np.power(vv-mid_points_uv[1][0,j+j_offset],2))
                        # print('u', uu- mid_points_uv[0][i+i_offset,0], r , grid_kernel_r_pixels/pixels_per_wavelength)
                        # print('v', vv- mid_points_uv[1][0,j+j_offset], r , grid_kernel_r_pixels/pixels_per_wavelength)
                        n_arr[j+j_offset, i+i_offset] += v_l * np.exp(-(r**2. / (grid_kernel_r_wavelength)**2.))
                        # print(i,j, i_offset, j_offset, r, v_l * np.exp(-(r**2. / (grid_kernel_r_pixels/pixels_per_wavelength)**2.)))
                i = uu_edges.__lt__(-uu).sum()-1
                j = vv_edges.__lt__(-vv).sum()-1
                for i_offset in offsets:
                    for j_offset in offsets:
                        r = np.sqrt(np.power(-uu-mid_points_uv[0][i+i_offset,0],2) + np.power(-vv-mid_points_uv[1][0,j+j_offset],2))
                        n_arr[j+j_offset, i+i_offset] += np.conjugate(v_l) * np.exp(-(r**2 / (grid_kernel_r_wavelength)**2))
            # apply the masked array and divide by number of entries
                mask = np.abs(n_arr).__gt__(0.)
                #mask = np.abs(n_arr).__gt__(vis_max_abs)
                n_arr[mask] = n_arr[mask]/np.abs(n_arr[mask])
        return (n_arr, uu_edges, vv_edges)


    def get_ift_simp(self, nw = 30, num_bin = 2**7):
        uv_plane = self.get_uvplane_zenith(num_bin=num_bin, nw=nw)
        ift = np.fft.fftshift(fft.ifft2(np.fft.ifftshift(uv_plane)))
        maxang = get_max_ang(nw, num_bin)
        extent = [maxang, -maxang, -maxang, maxang]
        return [ift, extent]


    def get_ift(self, nw = 30, num_bin = 2**7, use_kernel=False):
        uv_plane, uu_edges, vv_edges = self.get_uvplane(num_bin=num_bin, nw=nw, use_kernel=use_kernel)
        ift = np.fft.fftshift(fft.ifft2(np.fft.ifftshift(uv_plane)))
        maxang = get_max_ang(nw, num_bin)
        extent = [maxang, -maxang, -maxang, maxang]
        return [ift, extent]

    def get_beam(self, nw = 30, num_bin = 2**7, use_kernel=False):
        uv_plane, uu_edges, vv_edges = self.get_uvplane(num_bin=num_bin, nw=nw, use_kernel=use_kernel)
        ift = np.fft.ifftshift(fft.ifft2(np.fft.ifftshift(np.abs(uv_plane).__gt__(0))))
        return ift # /np.sum(ret)

    def get_image(self, CAL_IFT, CAL_EXTENT):
        abs_ift = np.abs(CAL_IFT)
        ift_std = np.std(abs_ift)
        ift_scaled = abs_ift/ift_std

        plt.figure(figsize=(8,6))
        plt.imshow(ift_scaled, extent=CAL_EXTENT)
        plt.colorbar()
        ift_median = np.median(abs_ift)
        ts = self.cal_vis_list[0].get_timestamp()
        plt.title( ts.strftime("%d-%m-%y %H:%M:%S") + ' std %1.3e median: %1.3e' % (ift_std, ift_median))
        plt.grid()
        plt.xlim(63, -63)
        plt.ylim(-63, 63)
        plt.xlabel('East-West')
        plt.ylabel('North-South')
        plt.tight_layout()



#!/usr/bin/python
# Simulate the known radio sources during an entire day's observing 
# Calculate visibilities and compare them to the observed visibilities
#
# Copyright (c) Tim Molteno 2013-2019. tim@elec.ac.nz
#
import traceback
import argparse
import math
import utc

import numpy as np
from multiprocessing import Pool


import datetime

import matplotlib.pyplot as plt
from tart.util import angle

import forward_map

from .visibility import *
from . import location
from ..operation import observation


  
def process_file(dirname, fname, orientation_degrees, spacing, sample_duration, simulate_vis=True):
  try:
    fpath = os.path.join(dirname, fname)
    nobs = observation.Observation_Load(fpath)
    date = nobs.timestamp.date()
    sec =  (nobs.timestamp - datetime.datetime.combine(date, datetime.time.min)).total_seconds()
    
    
    if (simulate_vis == True):
      utc_date = nobs.timestamp.replace(tzinfo=utc.UTC())
      antenna_locations = antenna_locations=nobs.ant_positions
    
      sky, baselines, visibility = forward_map.forward_map(utc_date, location.Dunedin, horizon=5.0, antenna_locations=antenna_locations, sample_duration=sample_duration) # Remove things below an elevation of five degrees
    else:
      sky = False
      baselines = False
      visibility = [False]
      
    hours = (sec / 3600.0)
    v = Visibility(nobs)
    return date, hours, v, visibility[0]
  except Exception as inst:
    print(type(inst))     # the exception instance
    print(inst.args)      # arguments stored in .args
    print(inst)
    print(sample_duration)
    print(traceback.format_exc())
    return None, None, None, None
  except KeyboardInterrupt:
    print(inst)
    return None, None, None, None

def get_correlation(dirname, orientation_degrees, spacing, sample_duration, simulate_vis=True):
  files = sorted(os.listdir(dirname))
  vis = []
  simvis = []
  times = []
  
  if (True):
    p = Pool()
  
    resultList = []

    try:
      for fname in files:
        if "data.pkl" in fname:
          resultList.append(p.apply_async(process_file, (dirname, fname, orientation_degrees, spacing, sample_duration, simulate_vis)))
      
      p.close 
      p.join
        
      for sv_thread in resultList:
        [date, hours,v,sv] = sv_thread.get() # timeout=2000
        if (date != None):
          times.append(hours)
          vis.append(v)
          simvis.append(sv)
          #print("%g %g %g" % (times[-1], vis[-1], simvis[-1]))
      p.terminate()
    except KeyboardInterrupt:
      print('control-c pressed')
      p.terminate()

  else:
    for fname in files:
      if "data.pkl" in fname:
        [date, hours,v,sv] = process_file(dirname, fname, orientation_degrees, spacing, sample_duration, simulate_vis)
        
        times.append(hours)
        vis.append(v)
        simvis.append(sv)
        
        print("%g %s %s" % (times[-1], v.v, simvis))
      
  #vis = np.array(vis)
  simvis = np.array(simvis)  
  return date, times, vis, simvis
  
if __name__ == '__main__':
  parser = argparse.ArgumentParser(description='Simulate the Sky and calculate simulated visibilities.')
  parser.add_argument('--data-dir', required=True, help="The filesystem path for the telescope data.")
  parser.add_argument('--orientation-angle', required=True, type=float, default=-22.0, help="The orientation of the array (degrees).")
  parser.add_argument('--optimize', action="store_true", help="Optimize the correlation between the simulated and measured visibilities.")
  parser.add_argument('--spacing', type=float, default=0.5, help="The spacing between antennas.")
  parser.add_argument('--sample-duration', type=float, default=1e-4, help="The length of time to simulate data for.")
  args = parser.parse_args()
  dirname = args.data_dir
  sample_duration = args.sample_duration
  orientation_degrees = args.orientation_angle
  spacing = args.spacing
  
  if (args.optimize):
    for theta in np.arange(orientation_degrees - 10.0, orientation_degrees+ 10.0, 1.0):
      for r in np.arange(spacing - 0.03, spacing + 0.03, 0.01):
    
        date, times, vis, simvis = get_correlation(dirname, theta, r, sample_duration)
        corr = sum(vis * simvis)
        print("Correlation @%f %f %f" % (theta, r, corr))
  else:
    date, times, vis, simvis = get_correlation(dirname, orientation_degrees, spacing, sample_duration)
    corr = sum(vis * simvis)
    print("Correlation @%f %f %f" % (orientation_degrees, spacing, corr))
   
    plt.subplot(2,1,1)
    title = "Visibility Data from %s (th=%f, dx=%f)" % (str(date), orientation_degrees, spacing)
    plt.title(title)
    plt.plot(times, vis,'.')
    plt.ylabel("Measured V")
    plt.xlabel("Hours since UTC ")
    plt.subplot(2,1,2)
    plt.plot(times, simvis,'.')
    plt.ylabel("Simulated V")
    plt.xlabel("Hours since UTC ")
    plt.savefig(os.path.join(dirname, "simulated_visibility_%f_%f.pdf" % (orientation_degrees, spacing)))
    plt.show()




'''
    Antenna Models for TART
    
    Author: Tim Molteno 2014-2019 tim@elec.ac.nz
    Copyright GPLv3
'''
import numpy as np
import healpy as hp

from tart.imaging import location
from tart.util import angle
from tart.util.db import db_connect, db_remove_antenna_measurements, db_insert_row



class AntennaModel(object):
    '''Base class for all Antenna models.'''

    def get_gain(self, el, az):
        '''Dummy method. All base classes should override this'''
        raise RuntimeError("You should be using a subclass of AntennaModel")

    def get_gain_equatorial(self, loc, utc_time, ra, decl):
        '''Return the gain from equatorial co-ordinates'''
        el, az = loc.equatorial_to_horizontal(utc_time, ra, decl)
        return self.get_gain(el, az)

class IdealHemisphericalAntenna(AntennaModel):
    '''Ideal Hemispherical Antenna.'''

    '''Return the gain '''
    def get_gain(self, el, az):
        if (el.to_degrees() < 0.0):
            return 0.0
        if (el.to_degrees() > 90):
            return 0.0
        return 1.0

class GpsPatchAntenna(AntennaModel):
    '''A more realistic model of a GPS patch antenna.'''

    def get_gain(self, el, az):
        if (el.to_degrees() < 5.0):
            return 0.0
        if (el.to_degrees() > 90):
            return 0.0
        return 1.0

def hp_interpolator(map_, el, az, n_pix=4):
    NSide = hp.pixelfunc.get_nside(map_)
    direction = np.array([np.pi/2.-el.to_rad(),az.to_rad()])
    steplength = hp.pixelfunc.max_pixrad(NSide)
    for i, r in enumerate(np.arange(steplength, np.pi, steplength)):
        pixels = np.array(hp.query_disc(NSide,hp.ang2vec(direction[0], direction[1]), r))
        filled = np.where(map_[pixels] > -1.)[0]
        l = len(filled)
        if l >= n_pix:
            # print(i, l)
            filled_pixel = pixels[filled]
            filled_pixel_directions = hp.pix2vec(NSide, filled_pixel)
            angular_distance = hp.rotator.angdist(direction, filled_pixel_directions)
            if angular_distance.min() == 0.: # do we really want this?
                return map_[filled_pixel[angular_distance.argmin()]]
            return np.average(map_[filled_pixel], weights=np.power(1./angular_distance, 2))

#def hp_spheric_harmonics(hp_alm, el, az, mmax=-1):
        #from scipy.special import sph_harm
        #if mmax==-1:
            #lmax = hp.Alm.getlmax(len(hp_alm))
            #mmax = lmax
        #else:
            #lmax = hp.Alm.getlmax(len(hp_alm),mmax=mmax)

        #print(lmax, mmax)
        #out = 0.
        #azimutal_angle = az.to_rad()
        #polar_angle = el.to_rad()-np.pi/2.
        #for l in range(lmax+1):
            #out += np.real(hp_alm[hp.Alm.getidx(lmax,l,0)] * sph_harm(0, l, azimutal_angle, polar_angle))
            #for m in range(1, 1+min(l , mmax)):
                #Ylm = sph_harm(m, l, azimutal_angle, polar_angle)
                #out += np.real(hp_alm[hp.Alm.getidx(lmax,l,m)] * Ylm)
                #out += np.real(hp_alm[hp.Alm.getidx(lmax,l,m)] * Ylm.conj()/np.power(-1.,m))
        #return out

def gen_interpolation_map(points, values, antenna_num, nside_exp):
    nside = np.power(2, nside_exp)
    npix = hp.nside2npix(nside)
    hp_map_avg = np.ones(npix) * hp.UNSEEN
    theta = np.array([90.-i[0] for i in points]) * np.pi/180.
    phi =     np.array([i[1] for i in points]) * np.pi/180.

    pix = hp.pixelfunc.ang2pix(nside, theta, phi)

    values = np.array(values)

    dic = {}
    for i, p in enumerate(pix):
        if dic.get(p) is None:
            dic[p] = []
        dic[p].append(values[i])

    pixel_dict = {}
    for p in dic:
        pixel_dict[p] = np.array(dic[p]).mean()
        hp_map_avg[p] = pixel_dict[p]

    map_mean = np.array([pixel_dict[key] for key in pixel_dict]).mean()

    for p in dic:
        #pixel_dict[p] = (pixel_dict[p]-map_mean)*(1.*len(hp_map_avg))/(4.*np.pi)
        hp_map_avg[p] = pixel_dict[p]

    return (hp_map_avg, pixel_dict)







class EmpiricalAntenna(AntennaModel):
    '''An antenna model that has a radiation pattern built up
         from measured GPS signal correlation amplitudes. This
         incorporates an horizon model as well.

         The model can be serialized into JSON form using the to_json()
         method. And reconstructed from JSON using the from_json()
    '''

    def __init__(self, antenna_num):
        self.sv = []        # A list of sv numbers - essentially an identifier
        self.distances = [] # A list of sv distances - essentially an identifier
        self.points = []    # The data is stored as a list of el, az.
        self.values = []    # The data is stored as a list of amplitudes.
        self.recalculate = True
        self.antenna_num = antenna_num
        self.interpolation_cache = []

    def add_measurement(self, el, az, amplitude, sv=-1, distance=-1):
        self.points.append([el.to_degrees(), az.to_degrees()])
        self.values.append(amplitude)
        self.sv.append(sv)
        self.distances.append(distance)
        self.recalculate = True

    def get_gain(self, el, az, n_pix=4, nside_exp_grid=10, nside_exp_syn=10, lmax=4, mmax=4, interpolate='default'):
        if (self.recalculate):
            self.recalculate = False
            self.i_map, self.pixel_dict = gen_interpolation_map(self.points, self.values, self.antenna_num, nside_exp_grid)
            if interpolate=='increasing_neighborhood':
                self.interp_gain = lambda el, az: hp_interpolator(self.i_map, el, az, n_pix)

                interp_gain()
                np.power(2, nside_exp_syn)
                self.norm_map = self.i_map
            else:
                alms = hp.sphtfunc.map2alm(hp.ma(self.i_map), lmax=lmax, mmax=mmax, iter=10, pol=False)
                print(hp.Alm.getlmax(len(alms), mmax=mmax))
                norm_map = hp.sphtfunc.alm2map(alms, np.power(2, nside_exp_syn), lmax=lmax, mmax=mmax, pixwin=False, pol=False)
                norm_map_min = norm_map.min()
                norm_map = norm_map-norm_map_min
                norm_map_max = norm_map.max()
                norm_map = norm_map/norm_map_max
                self.norm_map = norm_map
                # self.interp_gain = lambda el, az: (hp_spheric_harmonics(alms, el, az, mmax=mmax)-norm_map_min)/norm_map_max
                self.interp_gain = lambda el, az: self.norm_map[hp.pixelfunc.ang2pix(np.power(2, nside_exp_syn), np.pi/2-el.to_rad(), az.to_rad())]
        return self.interp_gain(el, az)

    def to_json(self, filename):
        import json
        ret = json.dumps({'antenna_num': self.antenna_num, 'sv': self.sv, 'distances': self.distances, 'points': self.points, 'values': self.values}, indent=4, separators=(',', ': '))
        f = open(filename, 'w')
        f.write(ret)
        f.close()

    def to_db(self, utc_date, db_file=None, table='gps_signals'):
        conn, sql = db_connect(db_file, table=table)
        
        db_remove_antenna_measurements(conn, sql, utc_date, self.antenna_num, table=table)
        cursor = conn.cursor()
        for p, v, sv, distance in zip(self.points, self.values, self.sv, self.distances):
            el, az = p
            correlation = v
            db_insert_row(cursor, sql, utc_date, self.antenna_num, sv, el, az, correlation, distance, table=table)

        conn.commit()
        conn.close()

    @classmethod
    def from_json(self, filename):
        import json
        f = open(filename, 'r')
        data = f.read()
        x = json.loads(data)
        f.close()
        ret = EmpiricalAntenna(x['antenna_num'])
        ret.points = x['points']
        ret.values = x['values']
        ret.sv = x['sv']
        ret.distances = x['distances']
        ret.recalculate = True
        return ret

    @classmethod
    def from_db(self, antenna_num, db_file=None, table='gps_signals'):
        ret = EmpiricalAntenna(antenna_num)
        conn, sql = db_connect(db_file, table=table)
        c = conn.cursor()
        c.execute(sql("SELECT el, az, correlation, date, sv, distance FROM "+table+" WHERE (antenna=%(ph)s)"), (antenna_num, ))
        #c.execute(sql("SELECT el, az, correlation, date FROM gps_signals WHERE (antenna=%(ph)s) AND el>15 AND correlation<6 AND date<%(ph)s"), (antenna_num, "2013-11-25"))
        # c.execute(sql("SELECT el, az, correlation, date FROM gps_signals WHERE (antenna=%(ph)s) AND date<%(ph)s"), (antenna_num, "2013-11-25"))
        pval = c.fetchall()
        print("Database has {} entries".format(len(pval)))
        points = []
        sv = []
        values = []
        times = []
        distances = []
        for p in pval:
            if ~np.isnan(p[2]):
                points.append((p[0],p[1]))
                values.append(p[2])
                times.append(p[3])
                sv.append(p[4])
                distances.append(p[5])
        ret.times = times
        ret.points = points
        ret.values = values
        ret.sv = sv
        ret.distances = distances
        ret.recalculate = True
        conn.close()
        return ret



# Copyright (C) Tim Molteno 2008-2019. All rights reserved

import math
from tart.simulation.util import Util

class Ephemeris(object):
    GM = 3.986005e14    # earth's universal gravitational parameter m^3/s^2
    WGS84_EARTH_ROTATION_RATE = 7.2921151467e-5;    # earth rotation rate, rad/s

    def __init__(self, in_hash):
        self.a0, self.a1, self.a2, self.a3 = in_hash['a'] # alpha parameters for ionospheric model
        self.b0, self.b1, self.b2, self.b3 = in_hash['b'] # beta parameters for ionospheric model

        self.svprn = in_hash['svprn']
        self.af0 = in_hash['af0']
        self.af1 = in_hash['af1']
        self.af2 = in_hash['af2']

        self.toc = in_hash['toc'] # GpsTime

        self.iode = in_hash['iode']
        self.crs = in_hash['crs']
        self.deltan = in_hash['deltan']
        self.m0 = in_hash['m0']

        self.cuc = in_hash['cuc']
        self.ecc = in_hash['ecc']
        self.cus = in_hash['cus']
        self.roota = in_hash['roota']

        self.toe = in_hash['toe']
        self.cic = in_hash['cic']
        self.omega_c = in_hash['omega_c']
        self.cis = in_hash['cis']

        self.i0 = in_hash['i0']
        self.crc = in_hash['crc']
        self.omega = in_hash['omega']
        self.omegadot = in_hash['omegadot']

        self.idot = in_hash['idot']
        self.codes = in_hash['codes']
        self.weekno = in_hash['weekno']
        self.l2flag = in_hash['l2flag']

        self.svaccur = in_hash['svaccur']
        self.svhealth = in_hash['svhealth']
        self.tgd = in_hash['tgd']

        self.tom = in_hash['tom']
        self.fit = in_hash['fit']
        self


    def from_hash(self, in_hash):
        ret = Ephemeris()
        ret.from_hash(in_hash)
        return ret


    def to_s(self):
        return "Ephemeris: SV={}, toe={}, toc={}, a0={}, ecc={}, m0={}, roota={}".format(self.svprn, self.toe, self.toc, self.a0, self.ecc, self.m0, self.roota)


    def to_hash(self):
        ret = {}
        ret['a']=[self.a0, self.a1, self.a2, self.a3] # alpha parameters for ionospheric model
        ret['b']=[self.b0, self.b1, self.b2, self.b3] # beta parameters for ionospheric model

        ret['svprn']=self.svprn
        ret['af0']=self.af0
        ret['af1']=self.af1
        ret['af2']=self.af2

        ret['toc']=self.toc # GpsTime

        ret['iode']=self.iode
        ret['crs']=self.crs
        ret['deltan']=self.deltan
        ret['m0']=self.m0

        ret['cuc']=self.cuc
        ret['ecc']=self.ecc
        ret['cus']=self.cus
        ret['roota']=self.roota

        ret['toe']=self.toe
        ret['cic']=self.cic
        ret['omega_c']=self.omega_c
        ret['cis']=self.cis

        ret['i0']=self.i0
        ret['crc']=self.crc
        ret['omega']=self.omega
        ret['omegadot']=self.omegadot

        ret['idot']=self.idot
        ret['codes']=self.codes
        ret['weekno']=self.weekno
        ret['l2flag']=self.l2flag

        ret['svaccur']=self.svaccur
        ret['svhealth']=self.svhealth
        ret['tgd']=self.tgd

        ret['tom']=self.tom
        ret['fit']=self.fit
        return ret


    #\brief Find the clock correction from UTC for this SV
    #         This is done ignoring leap seconds. In other words, this is modulo 1 second.
    #
    def clock_correct(self, sow):
        # more accurate calculation
        e = self.getE(self.Tsv)
        dt = Util.check_t(self.Tsv - self.toc.sow())
        dtr = -4.442807e-10 * self.ecc * self.roota * math.sin(e)     #relatavistic correction
        dtsv = self.af0 + self.af1*(dt) + self.af2*dt*dt + dtr - self.tgd
        return dtsv


    def get_tk(self, sow):
        tk = Util.check_t(sow-self.toe) #        Time from ephemeris reference epoch (1)
        if (tk < -302400.0):
            raise RuntimeError("Invalid time %f" % tk)
        if (tk > 302400.0):
            raise RuntimeError("Invalid time %f" % tk)
        return tk

    # http://home-2.worldonline.nl/~samsvl/satpos.htm
    # Get the eccentricity
    def getE0(self, sow):

        x=self.m0             # kepler's equation for eccentric anomaly ek
        y=self.m0 - (x-self.ecc*math.sin(x))
        x1=x
        x=y
        for i in range(0,15):
            x2=x1
            x1=x
            y1=y
            y = self.m0 - (x- self.ecc*math.sin(x))
            if (abs(y-y1)<1.0e-15):
                break
            x=(x2*y-x*y1)/(y-y1)
        ek=x        # end of det. of ecc. anomaly

        tk = self.get_tk(sow)
        n0 = math.sqrt(Ephemeris.GM/(self.roota**6)) #        Computed mean motion
        n = n0+self.deltan                        #        Corrected mean motion
        mk=self.m0+n*tk    #    mean anomaly

        x=mk    #    kepler's equation for eccentric anomaly ek
        y=mk-(x- self.ecc* math.sin(x))
        x1=x
        x=y
        for i in range(0,15):
            x2=x1
            x1=x
            y1=y
            y=mk-(x- self.ecc* math.sin(x))
            if(abs(y-y1)<1.0E-15):
                break
            x=(x2*y-x*y1)/(y-y1)
        ek=x

        print(("    E0 -> {}".format(ek)))
        return ek

    def getE(self, sow):
        a = self.roota*self.roota
        tk = self.get_tk(sow)
        n0 = math.sqrt(Ephemeris.GM/(a**3)) #  Computed mean motion
        n = n0+self.deltan                  #  Corrected mean motion
        m = self.m0+n*tk                    #  Mean anomaly

        #test = self.getE0(sow)
        m = Util.rem2pi(m + Util.PI2)
        e = m
        for i in range(0,15):
            e_old = e
            e = m + self.ecc*math.sin(e_old)
            dE = Util.rem2pi(e - e_old)
            if (abs(dE) < 1.0e-15):
                break

        e = Util.rem2pi(e + Util.PI2)
        return e


    def get_sv_position(self, gt):
        return self.get_location(gt.sow())

    def get_sv_position_utc(self, utc_datetime):
        gpst = gps_time.GpsTime.from_time(utc_datetime)
        return self.get_location(gpst.sow())

    def get_location(self, sow):
        a = self.roota*self.roota        # Semi major axis
        tk = self.get_tk(sow) #     tk = sow-@toe


        e = self.getE(sow)

        v = math.atan2(math.sqrt(1.0-(self.ecc**2))*math.sin(e), math.cos(e)-self.ecc)
        phi = v+self.omega
        phi = Util.rem2pi(phi)
        phi2 = 2.0*phi

        cosphi2 = math.cos(phi2)
        sinphi2 = math.sin(phi2)

        u = phi + self.cuc*cosphi2+self.cus*sinphi2
        r = a*(1.0-self.ecc*math.cos(e)) + self.crc*cosphi2+self.crs*sinphi2
        i = self.i0+self.idot*tk + self.cic*cosphi2+self.cis*sinphi2
        om = self.omega_c + (self.omegadot - Ephemeris.WGS84_EARTH_ROTATION_RATE)*tk - \
            Ephemeris.WGS84_EARTH_ROTATION_RATE*self.toe
        om = Util.rem2pi(om + Util.PI2)
        print(("w_c={}, wdot={}, om={}".format(self.omega_c, self.omegadot, om)))
        x1 = math.cos(u)*r
        y1 = math.sin(u)*r

        x = x1*math.cos(om) - y1*math.cos(i)*math.sin(om)
        y = x1*math.sin(om) + y1*math.cos(i)*math.cos(om)
        z = y1*math.sin(i)
        print(("ephemeris.get_location({}) tk={}, {} -> x:{}".format(sow, tk, self.to_s(), x)))
        return [x,y,z]

    
    def get_location_new(self, sow):
        ''' A sanity check from a different source
            http://gnsstk.sourceforge.net/gps_8c-source.html
            https://gssc.esa.int/navipedia/index.php/GPS_and_Galileo_Satellite_Coordinates_Computation
        '''
        a = self.roota*self.roota        # Semi major axis
        tk = Util.check_t(sow - self.toe)

        Mk =  self.m0 + (math.sqrt(Ephemeris.GM/(a**3)) + self.deltan)*tk

        E_0 = Mk
        for _ in range(7):
            E_k = Mk + self.ecc*math.sin(E_0)
            E_0 = E_k

        v_k = math.atan(math.sqrt(1.0-(self.ecc**2))*math.sin(E_k)/math.cos(E_k)-self.ecc)

        phi = v_k+self.omega
        phi2 = 2.0*phi

        cosphi2 = math.cos(phi2)
        sinphi2 = math.sin(phi2)

        u_k = phi + self.cuc*cosphi2 + self.cus*sinphi2
        r_k = a*(1.0 - self.ecc*math.cos(E_k)) + self.crc*cosphi2 + self.crs*sinphi2
        i_k = self.i0 + self.idot*tk + self.cic*cosphi2 + self.cis*sinphi2
        
        lambda_k = self.omega_c + (self.omegadot - Ephemeris.WGS84_EARTH_ROTATION_RATE)*tk - \
            Ephemeris.WGS84_EARTH_ROTATION_RATE*self.toe

        x1 = math.cos(u_k)*r_k
        y1 = math.sin(u_k)*r_k

        x = x1*math.cos(lambda_k) - y1*math.cos(i_k)*math.sin(lambda_k)
        y = x1*math.sin(lambda_k) + y1*math.cos(i_k)*math.cos(lambda_k)
        z = y1*math.sin(i_k)
        return [x,y,z]

    def get_velocity(self, sow):
        loc1 = np.array(self.get_location(sow - 0.5))
        loc2 = np.array(self.get_location(sow + 0.5))
        return (loc1 - loc2)



import unittest

from tart.simulation.util import *
import numpy as np

class TestUtil(unittest.TestCase):

    def test_rad2deg(self):
        self.assertEqual(Util.rad2deg(0.0), 0.0)
        self.assertEqual(Util.rad2deg(2*Util.PI), 360.0)
        self.assertEqual(Util.rad2deg(Util.PI), 180.0)

    def test_rem(self):
        for k in range(1000):
            x = np.random.uniform(-1000,1000)
            y = np.random.uniform(-100,100)
            self.assertAlmostEqual(Util.rem(x,y), x % y, 9)

    def test_rem2pi(self):
        self.assertEqual(Util.rem2pi(Util.PI), Util.PI)
        self.assertEqual(Util.rem2pi(Util.PI+Util.PI2), Util.PI)
        self.assertEqual(Util.rem2pi(Util.PI-Util.PI2), Util.PI)
        self.assertEqual(Util.rem2pi(Util.PI/10-Util.PI2), Util.PI/10)

    def test_idiv(self):
        for k in range(1000):
            x = np.random.uniform(-1000,1000)
            y = np.random.uniform(-100,100)
            self.assertEqual(Util.idiv(x,y), x//y)




# from tart.imaging import radio_source
# from tart.imaging.ephemerides_proxy import EphemeridesProxy
from tart.imaging import location
# from tart.util import utc
# from tart.util import constants
# from tart.util import vector

import datetime
import numpy as np

import unittest
from tart.util import utc
from tart.util import angle

from tart.imaging.gps_satellite import *

class TestGpsSatellite(unittest.TestCase):

  def setUp(self):
    self.t = utc.utc_datetime(2002, 10, 31, 0, 2, 2)

  def test_gps(self):

    import ephem
    dnd = ephem.Observer()
    dnd.lon = str(location.Dunedin.lon.to_degrees())
    dnd.lat = str(location.Dunedin.lat.to_degrees())
    dnd.elevation = location.Dunedin.alt
    dnd.pressure = 0.0
    t = utc.utc_datetime(2013,9,20,0,0,3)
    dnd.date = t


    for svnum in range(31,32):
      l1 = 'GPS BIIRM-2 (PRN 31)    '
      l2 = '1 29486U 06042A   13264.26023969 -.00000084  00000-0  00000+0 0  2292'
      l3 = '2 29486  56.1466 336.5502 0081195 316.5190  69.5394  2.00563866 51246'
      j = ephem.readtle(l1,l2,l3)
      # http://blog.thetelegraphic.com/2012/gps-sattelite-tracking-in-python-using-pyephem/
      #j = ephem.Gps(svnum)
      j.compute(dnd)
      ra_e = angle.from_rad(j.ra.real)
      dec_e= angle.from_rad(j.dec.real)

      sv = GpsSatellite(svnum, location.Dunedin)
      ra, dec = sv.radec(t)

      print(angle.wrap_360(ra.to_degrees()))
      self.assertAlmostEqual(ra.to_degrees(), ra_e.to_degrees(), 0)
      self.assertAlmostEqual(dec.to_degrees(), dec_e.to_degrees(), 0)

  def test_radec(self):
    sv = GpsSatellite(31, location.Dunedin)
    t = utc.utc_datetime(2013,9,20,0,0,3)
    ra, dec = sv.radec(t)
    self.assertLess(ra.to_degrees(), 360.0)
    self.assertLess(abs(dec.to_degrees()), 90.0)

  def test_power(self):
    sv = GpsSatellite(31, location.Dunedin)
    t = utc.utc_datetime(2013,9,20,0,0,3)
    self.assertLess(sv.jansky(t), 5.0e6)
    self.assertGreater(sv.jansky(t), 3.0e6)



'''
Generate UVFITS file from visibility object.
'''

from astropy.io import fits
import numpy as np

from tart.imaging import tart_util
from tart.imaging import radio_source
from tart.imaging import location

from tart.simulation import antennas

from tart.util import angle
from tart.util import skyloc
from tart.util import constants


def encode_baseline(a1, a2):
    '''Encode the baseline. Use the miriad convention to handle more than 255 antennas (up to 2048).'''
    if a1 > a2:
        te = a2
        a2 = a1
        a1 = te

    if (a2 > 255):
        return a1*2048 + a2 + 65536 # this is backwards compatible with the standard UVFITS convention
    else:
        return a1*256 + a2


def decode_baseline(bl):
    '''Decode a baseline using same (extended) miriad format as above'''
    if (bl > 65536):
        bl -= 65536
        a2 = bl % 2048
        a1 = (bl - a2)/2048
        return a1 , a2
    else:
        a2 = bl % 256
        a1 = (bl - a2)/256
        return a1 , a2


class UVFitsGenerator(object):
    ''' Generate a UVFITS file for a target source from an array of visibilities.'''
    def __init__(self, cal_vis_list, phase_center): # FIXME
        # self.phase_center_elevation = phase_center_elevation
        # self.phase_center_azimuth = phase_center_azimuth
        vt = cal_vis_list[0]

        c = vt.get_config()
        ant_p = np.asarray(c.get_antenna_positions())

        self.config = c
        self.v_array = cal_vis_list
        self.n_baselines = len(vt.get_baselines())
        self.loc = c.get_loc() # Assume the array isn't moving
        
        if phase_center is None:
                ra, dec = location.get_loc(vt.get_config()).horizontal_to_equatorial(vt.get_timestamp(), angle.from_dms(90), angle.from_dms(0))
                self.phase_center = radio_source.CosmicSource(ra, dec, 1e10)
        else:
                self.phase_center = phase_center

    def gen_vis_table(self):
        '''Generate visibility table and fill payload.'''

        baselines = []
        ant_pos = self.config.get_antenna_positions()
        loc = self.config.get_loc()
        for cal_vis in self.v_array:
            ra, dec = self.phase_center.radec(cal_vis.vis.timestamp)
            #v.rotate(skyloc.Skyloc(ra, dec))
            uu_a, vv_a, ww_a = cal_vis.get_all_uvw()
            bls = cal_vis.get_baselines()
            datestamp = tart_util.get_julian_date(cal_vis.vis.timestamp)- int(tart_util.get_julian_date(cal_vis.vis.timestamp)+0.5)
            for uu,vv,ww, b in zip(uu_a, vv_a, ww_a, bls):
                baseline = {}
                [i,j] = b
                # print((np.array(a1.enu) - np.array(a0.enu)), uu, vv, ww)
                # arcane units of UVFITS require u,v,w in nanoseconds
                baseline['UU'] = uu*constants.L1_WAVELENGTH/constants.V_LIGHT
                baseline['VV'] = vv*constants.L1_WAVELENGTH/constants.V_LIGHT
                baseline['WW'] = ww*constants.L1_WAVELENGTH/constants.V_LIGHT
                baseline['BASELINE'] = encode_baseline(i+1, j+1)
                baseline['DATE'] = datestamp
                # DATE FIXME ?
                baselines.append(baseline)

        freqs = np.array([1545.])
        pols = np.array(['+'])

        data = np.zeros((len(self.v_array)*self.n_baselines, 1, 1, len(freqs), len(pols), 3))

        for i, v in enumerate(self.v_array):
            for k,    b in enumerate(cal_vis.get_baselines()):
                for l, _ in enumerate(freqs):
                    for j, _ in enumerate(pols):
                        vis    = v.get_visibility(b[0], b[1])
                        re = vis.real
                        img = vis.imag
                        w = np.ones(1)
                        data[i*self.n_baselines+k, 0, 0, l, j, :] = [re, img, w]

        hdu = fits.GroupsHDU(fits.GroupData(data, parnames=['UU', 'VV', 'WW', 'BASELINE', 'DATE'], bitpix=-32, \
        pardata=[     [b['UU']                for b in baselines], \
                                [b['VV']                for b in baselines], \
                                [b['WW']                for b in baselines], \
                                [b['BASELINE']    for b in baselines], \
                                [b['DATE']            for b in baselines]]))
        return hdu


    def gen_ant_table(self):
        '''Generate antenna table.'''
        v = self.v_array[0]
        n_ant = self.config.get_num_antenna()
        ant_pos = self.config.get_antenna_positions()
        v.antennas = []
        for i in range(n_ant):
            antenna = {}
            antenna['ANNAME']    = 'ANT%i'% (i+1)         # there is no antenna 0
            antenna['STABXYZ'] = ant_pos[i]
            antenna['ORBPARM'] = 0. # SET TO 0. - not used, because 'NUMORB' in header is    0
            antenna['NOSTA']     = i+1                                # there is no antenna 0
            antenna['MNTSTA']    = 0
            antenna['STAXOF']    = 0.
            antenna['POLTYA']    = 'X'
            antenna['POLAA']     = [0.]
            antenna['POLCALA'] = np.array([0., 0., 0.])
            antenna['POLTYB']    = 'Y'
            antenna['POLAB']     = [0.]
            antenna['POLCALB'] = np.array([0., 0., 0.])
            v.antennas.append(antenna)

        c1 = fits.Column(name='ANNAME',     format='8A', array= [a['ANNAME']    for a in v.antennas])
        c2 = fits.Column(name='STABXYZ',    format='3D', array= [a['STABXYZ'] for a in v.antennas], unit='METERS') 
        c3 = fits.Column(name='NOSTA',        format='1J', array= [a['NOSTA']     for a in v.antennas])
        c4 = fits.Column(name='MNTSTA',     format='1J', array= [a['MNTSTA']    for a in v.antennas])
        c5 = fits.Column(name='STAXOF',     format='1E', array= [a['STAXOF']    for a in v.antennas], unit='METERS')
        c6 = fits.Column(name='POLTYA',     format='1A', array= [a['POLTYA']    for a in v.antennas])
        c7 = fits.Column(name='POLAA',        format='1E', array= [a['POLAA']     for a in v.antennas], unit='DEGREES')
        c8 = fits.Column(name='POLCALA',    format='3E', array= [a['POLCALA'] for a in v.antennas])
        c9 = fits.Column(name='POLTYB',     format='1A', array= [a['POLTYB']    for a in v.antennas])
        c10 = fits.Column(name='POLAB',     format='1E', array= [a['POLAB']     for a in v.antennas], unit='DEGREES')
        c11 = fits.Column(name='POLCALB', format='3E', array= [a['POLCALB'] for a in v.antennas])
        c0 = fits.Column(name='ORBPARM',    format='1D', array= [a['ORBPARM'] for a in v.antennas])

        return fits.BinTableHDU.from_columns([c1, c2, c3, c4, c5, c6, c7, c8, c9, c10, c11, c0])


    def update_vis_header(self):
        '''Update visibility header.'''

        v = self.v_array[0]
        params = ["UU", "VV", "WW", "BASELINE", "DATE"]
        ctypes = ["STOKES", "FREQ", "RA", "DEC"]

        self.vis_header.set('OBJECT', 'Full Sky')

        # without rotation of the visibilities is has to be this:
        # ra, dec = self.loc.horizontal_to_equatorial(v.vis.timestamp, angle.from_dms(90.), angle.from_dms(0.))
        # with rotation is has to be:

        ra, dec = self.phase_center.radec(v.vis.timestamp)

        self.vis_header.set('OBSRA', ra.to_degrees())	# Antenna pointing center
        self.vis_header.set('OBSDEC', dec.to_degrees())	# Antenna pointing center

        self.vis_header.set('TELESCOP', 'TART')
        self.vis_header.set('INSTRUME', 'TART')
        self.vis_header.set('EPOCH', 2000.0)
        self.vis_header.set('BSCALE', 1)
        self.vis_header.set('DATE-OBS', '%i-%02i-%02iT00:00:00.0'%(v.vis.timestamp.year, v.vis.timestamp.month, v.vis.timestamp.day))

        for i, par in enumerate(params):
            self.vis_header.set('PTYPE%i' % (i+1), par)
            self.vis_header.set('PSCAL%i' % (i+1), 1.0)
            self.vis_header.set('PZERO%i' % (i+1), 0.0)

        self.vis_header.set('PZERO5', int(tart_util.get_julian_date(v.vis.timestamp) + 0.5))

        self.vis_header.set('CRVAL2', 1.0)
        self.vis_header.set('CRPIX2', 1.0)
        self.vis_header.set('CDELT2', 1.0)
        self.vis_header.set('CTYPE2', 'COMPLEX')

        for i, ctyp in enumerate(ctypes):
            self.vis_header.set('CTYPE%d' % (i+3), ctyp)
            self.vis_header.set('CRVAL%d' % (i+3), 1.0)
            self.vis_header.set('CRPIX%d' % (i+3), 1.0)
            self.vis_header.set('CDELT%d' % (i+3), 1.0)

        #set the input pol type.
        #for circular CRVAL3 should be -1, CDELT3 -1
        #for linear     CRVAL3 should be -5, CDELT3 -1
        #for Stokes     CRVAL3 should be    1, CDELT3    1
        #header['CRVAL3']    = -5.
        #header['CDELT3']    = -1.

        pol_typ = 1.    # FIXME
        #pol_typ =    Obs.pol_typ
        self.vis_header.set('CRPIX3', pol_typ)
        if pol_typ < 0:
            self.vis_header.set('CDELT3', -1.)
        else:
            self.vis_header.set('CDELT3', 1.)

        self.vis_header.set('CRPIX4', 1.)    # something .... n_freq/2 +1
        self.vis_header.set('CDELT4', v.vis.config.get_bandwidth()) # freq delta
        self.vis_header.set('CRVAL4', v.vis.config.get_operating_frequency()) # center freq

        ra, dec = self.phase_center.radec(v.vis.timestamp)

        self.vis_header.set('CRVAL5', ra.to_degrees())	    #SOURCE RA    PARAMETER
        self.vis_header.set('CRVAL6', dec.to_degrees())	#SOURCE DEC PARAMETER

        self.vis_header.add_history("AIPS WTSCAL =    1.0")    # has to be in here for some reason
        self.vis_header.add_comment("written by the TART UV FITS writer. ELEC Otago Uni")

    def update_ant_header(self):
        '''Update antenna header.'''

        v = self.v_array[0]

        self.ant_header.set('TFORM3', '1J')    #
        self.ant_header.set('TFORM4', '1J')        #
        self.ant_header.set('TFORM5', '1E')         #     replace (J,E,D) with (1J,1E,1D) to match difmap requirements
        self.ant_header.set('TFORM7', '1E')         # # ra and dec of sun here! FIXME - move to fixed phase center in ra/dec
        self.ant_header.set('TFORM10', '1E')     #
        self.ant_header.set('TFORM12', '1D') #

        self.ant_header.set('EXTNAME','AIPS AN')
        self.ant_header.set('XTENSION', 'BINTABLE')
        arr_loc = self.loc.get_ecef()
        self.ant_header.set('ARRAYX', arr_loc[0]) #
        self.ant_header.set('ARRAYY', arr_loc[1])    #    TART POSITION in ECEF coordinates
        self.ant_header.set('ARRAYZ', arr_loc[2]) #
        self.ant_header.set('FREQ', v.vis.config.get_operating_frequency())
        #/* GSTIAO is the GST at zero hours in the time system of TIMSYS (i.e. UTC) */
        #mjd = trunc(data->date[0] - 2400000.5);
        #temp = slaGmst(mjd)*180.0/M_PI;    # technically, slaGmst takes UT1, but it won't matter here.
        #fits_update_key(fptr,TDOUBLE,"GSTIA0",&temp , NULL, &status);
        self.ant_header.set('GSTIA0', self.loc.GST(v.vis.timestamp).to_degrees())
        self.ant_header.set('DEGPDY', 3.60985642988E+02) # Rotation per day

        self.ant_header.set('RDATE', "%d-%02d-%02dT00:00:00.0"%(v.vis.timestamp.year, v.vis.timestamp.month, v.vis.timestamp.day))
        self.ant_header.set('POLARX', 0.0)
        self.ant_header.set('POLARY', 0.0)
        self.ant_header.set('UT1UTC', 0.0)
        self.ant_header.set('DATUTC', 0.0)
        self.ant_header.set('TIMSYS', 'UTC')
        self.ant_header.set('ARRNAM', 'TART')
        self.ant_header.set('NUMORB', 0)
        self.ant_header.set('NOPCAL', 3)
        self.ant_header.set('FREQID', -1)
        self.ant_header.set('IATUTC', 33.0)
        self.ant_header.set('EXTVER', 1)
        #self.ant_header.set('XYZHAND','RIGHT')

    def write(self, filename):
        '''Generate tables and headers and write UVFITS file.'''
        vis_table = self.gen_vis_table()
        ant_table = self.gen_ant_table()

        self.vis_header = vis_table.header
        self.update_vis_header()
        self.ant_header = ant_table.header
        self.update_ant_header()

        hdulist = fits.hdu.hdulist.HDUList([vis_table, ant_table])
        hdulist.writeto(filename)



#
# A class the encapsulates a location on earth. Does all the necessary frame-of-reference
# conversions.
#
# Copyright (c) Tim Molteno 2013-2017. tim@elec.ac.nz
#

import math
import numpy as np

from tart.util import angle
from tart.imaging import tart_util

def get_loc(Settings):
    return Location(angle.from_dms(Settings.get_lat()), angle.from_dms(Settings.get_lon()), Settings.get_alt())


''' Future class structure going in here'''
class ENU(object):
    def __init__(self, e, n, u):
        self.e = e
        self.n = n
        self.u = u

class Horizontal(object):
    def __init__(self, r, el, az):
        self.r = r
        self.el = el
        self.az = az

class ECEF(object):
    def __init__(self, x, y, z):
        self.x = x
        self.y = y
        self.z = z

class ECI(object):
    def __init__(self, x, y, z):
        self.x = x
        self.y = y
        self.z = z


# Convert ECI (Earth-Centered-Inertial) to ECEF coordinates.
def eci_to_ecef(utc_date, x_in, y_in, z_in):
    ''' cos(theta) -sin(theta)  0
        sin(theta)  cos(theta)  0
        0           0           1
    '''
    # Rotate x and y by the hour angle
    theta = -Location.GST(utc_date)

    return [x_in*theta.cos() - y_in*theta.sin(), y_in*theta.cos() + x_in*theta.sin(), z_in]

# Convert ECEF to ECI (Earth-Centered-Inertial)
def ecef_to_eci(utc_date, x_in, y_in, z_in):

    # Undo Rotate x and y by the hour angle
    theta = -Location.GST(utc_date)

    return [x_in*theta.cos() + y_in*theta.sin(), y_in*theta.cos() - x_in*theta.sin(), z_in]

class Location(object):
    R_EARTH = 6378137.0 # earth semimajor axis in meters
    F_RECIP = 1.0/298.257223563 # reciprocal flattening
    E2 = 2.0*F_RECIP - F_RECIP*F_RECIP # eccentricity squared

    def __init__(self, lat, lon, alt):
        self.lat = lat # requires init with lat and lon as angle objects (NEW)
        self.lon = lon # requires init with lat and lon as angle objects (NEW)
        self.alt = alt
        self.ex, self.ey, self.ez = self.calculate_ecef()

        
    def __repr__(self):
        return "[lat=%f, lon=%f, alt=%f]" % (self.latitude_deg(), self.longitude_deg(), self.alt)

    def longitude_deg(self):
        return angle.wrap_180(self.lon.to_degrees())

    def latitude_deg(self):
        return self.lat.to_degrees()

    def altitude(self):
        return self.alt

    def get_XYZ(self, e, n, u):
        '''Return the location in terms of the XYZ (non-earth fixed coordinate system).
             The X axis starts at the earths center and points through (HA = 0, decl = 0)
             The Y axis points at 90 degrees to X and in the plane of the equator (HA = -90, decl=0)
             Z points through the north celestial pole (decl = 90)'''
        ecef0 = self.get_ecef() # ECEF of origin
        # Now create a new location offset by ENU
        d_lon = angle.from_rad((e/self.R_EARTH)*self.lat.sin())
        d_lat = angle.from_rad(n/self.R_EARTH)

        loc2 = Location((self.lat + d_lat), (self.lon + d_lon), self.alt + u)
        ecef2 = loc2.get_ecef()

        return np.array(ecef2) - np.array(ecef0)

    def get_ecef_wgs84(self):
        a = 6378137.0 # m Semi-major axis
        b = 6356752.314245 # Semi-minor axis
        inverse_flattening = 298.257223563
        
        sinlat = self.lat.sin()
        n = a*a / (np.sqrt(a*a*self.lat.cos()**2 + b*b*sinlat**2))
        x = (n + self.alt)*self.lat.cos()*self.lon.cos()
        y = (n + self.alt)*self.lat.cos()*self.lon.sin()
        z = (n*(b**2 / a**2) + self.alt)*sinlat
        return [x, y, z]

    def calculate_ecef(self):
        ''' Position of the Location in ECEF coordinates '''
        sinlat = self.lat.sin()
        chi = math.sqrt(1.0 - Location.E2*sinlat*sinlat)
        coslat = self.lat.cos()
        achih = (Location.R_EARTH/chi + self.alt)
        x = achih*coslat*self.lon.cos()
        y = achih*coslat*self.lon.sin()
        z = (Location.R_EARTH*(1.0-Location.E2)/chi + self.alt)*sinlat

        return [x, y, z]

    def get_ecef(self):
        return [self.ex, self.ey, self.ez]

     # ECEF to ENU
    def ecef_to_enu(self, x_in, y_in, z_in):
        lon = self.lon
        lat = self.lat
        e = -x_in*lon.sin()           + y_in*lon.cos()
        n = -x_in*lon.cos()*lat.sin() - y_in*lon.sin()*lat.sin() + z_in*lat.cos()
        u =  x_in*lon.cos()*lat.cos() + y_in*lon.sin()*lat.cos() + z_in*lat.sin()
        return [e, n, u]

    def get_ecef_delta_from_enu(self, e_in, n_in, u_in):
        # FIXME: definitely TEST ENU_TO_ECEF
        lat = self.lat
        lon = self.lon
        x = -lon.sin() * e_in - lon.cos()*lat.sin()*n_in + lon.cos()*lat.cos()*u_in
        y =  lon.cos() * e_in - lon.sin()*lat.sin()*n_in + lon.sin()*lat.cos()*u_in
        z =  lat.cos() * n_in + lat.sin() * u_in
        return [x, y, z]

    # http://www.navipedia.net/index.php/Transformations_between_ECEF_and_ENU_coordinates
    def ecef_to_horizontal(self, x_in, y_in, z_in):

        ex,ey,ez = self.get_ecef() # My position in ECEF

        rx,ry,rz = [x_in - ex, y_in - ey, z_in - ez]
        enu = np.array(self.ecef_to_enu(rx,ry,rz))

        r = np.sqrt(enu.dot(enu))
        rho = enu / r

        el = angle.asin(rho[2])

        n = rho[1] #n
        e = rho[0] #e

        az = angle.atan2(e, n)
        if az.to_degrees() < 0.:
            az = angle.from_dms(360. + az.to_degrees())

        return [r, el, az]

    @staticmethod
    def horizontal_to_enu(r, el, az):
        ''' Get ENU from Horizontal (local Topocentric)
        '''
        e = el.cos()*az.sin()
        n = el.cos()*az.cos()
        u = el.sin()

        return [r*e, r*n, r*u]

    def horizontal_to_ecef(self, r, el, az):
        ''' Convert r,el,az to ECEF '''
        ## Convert to ENU
        e,n,u = Location.horizontal_to_enu(r, el, az)
        
        # Convert ENU to ECEF
        delta_ecef = self.get_ecef_delta_from_enu(e, n, u)
        my_ecef = np.array(self.get_ecef())

        return my_ecef + delta_ecef
        
    @staticmethod
    def GST(utc_date):
        ''' Return the Greenwich Sidereal Time'''
        JD = tart_util.JulianDay(utc_date)
        D = JD - 2451545.0
        T = int(D // 36525)

        GMST = 18.697374558 + 24.06570982441908*D

        # Now correct for precession of equioxes
        epsilon = 23.4393 - 0.0000004*D
        L = 280.47 + 0.98565*D
        omega = 125.04 - 0.052954*D
        dphi = -0.000319*math.sin(angle.deg_to_rad(omega)) - 0.000024*math.sin(angle.deg_to_rad(2.0*L))
        eqeq = dphi * math.cos(angle.deg_to_rad(epsilon))

        return angle.from_hours(GMST + eqeq)

    @staticmethod
    def GHA(utc_date, ra):
        ''' Calculate the Greenwich Hour Angle (GHA)'''
        return Location.GST(utc_date) - ra

    def LST(self, utc_date):
        ''' Calculate the Local Sidereal Time '''
        return Location.GST(utc_date) + self.lon

    def LHA(self, utc_date, ra):
        ''' Calculate the Local Hour Angle (LHA)'''
        return self.LST(utc_date) - ra

    def horizontal_to_LHA(self, utc_date, el, az):
        phi_0 = self.lat.to_rad()

        tan_h = az.sin() / (az.cos() * math.sin(phi_0) + el.tan() * math.cos(phi_0))
        local_hour_angle = angle.atan(tan_h)
        return local_hour_angle


    def equatorial_to_horizontal(self, utc_date, ra, dec):
        ''' Convert RA/Decl to Elevation Azimuth this location '''
        # [Peter Duffett-Smith, Jonathan_Zwart] Practical Astronomy with calculator and spreadsheet
        lha = self.LHA(utc_date,ra)
        lat = self.lat

        el = angle.asin(dec.sin()*lat.sin() + dec.cos()*lat.cos()*lha.cos())

        az = angle.atan2((-lha.sin())*dec.cos(), (dec.sin()-lat.sin()*el.sin())/lat.cos())

        if az.to_degrees() < 0.:
            az = angle.from_dms(360. + az.to_degrees())

        return el, az


    def horizontal_to_equatorial(self, utc_date, el, az):
        ''' Convert an azimuth and elevation to RA/Decl
                Useful for looking straight up, and working out the RA/Declination
                Return RA , Decl as angle objects
        '''
        lat = self.lat
        dec_sin = (el.sin() * lat.sin()) + (el.cos() * lat.cos() * az.cos())
        dec = angle.asin(dec_sin)

        LST = self.LST(utc_date)

        H = angle.atan2(-1.*el.cos()*lat.cos()*az.sin(), (el.sin()-(lat.sin()*dec_sin)))

        ra = LST - H

        return ra.to_ra(), dec.to_declination()

    def horizontal_to_eci(self, r, el, az, utc_date):
        x, y, z = self.horizontal_to_ecef(r, el, az)
        xi, yi, zi = ecef_to_eci(utc_date, x, y, z)
        return [xi, yi, zi]

    def eci_to_horizontal(self, x, y, z, utc_date):
        xf, yf, zf = eci_to_ecef(utc_date, x, y, z)
        r, el, az = self.ecef_to_horizontal(xf, yf, zf)
        return [r,el,az]

'''Convenient helper function for the location of the Physics Department Roof'''
Dunedin = Location(lat=angle.from_dms(-45.86391200), lon=angle.from_dms(170.51348452), alt=46.5)

'''Convenient helper function for the location of the rural TART'''
Dunedin_Farm = Location(lat=angle.from_dms(-45.851868), lon=angle.from_dms(170.545558), alt=266.5)

'''Convenient helper function for somewhere cold and damp in the far north'''
Aachen = Location(lat=angle.from_dms(50.778), lon=angle.from_dms(6.086), alt=46.5)



import json
import h5py
import dateutil
import os

import numpy as np

try:
   import cPickle as pickle
except:
   import pickle

from tart.util import angle
from tart.util import skyloc
from tart.util import constants

from tart.operation import observation
from tart.operation import settings

from tart.simulation.simulation_source import HorizontalSource

class Visibility:
    """
    A container class for visibilities from a single observation.
    """
    def __init__(self, obs, phase_el, phase_az):
        self.phase_el = phase_el
        self.phase_az = phase_az
        self.config = obs.config
        self.timestamp = obs.timestamp

    def __init__(self, config, timestamp):
        self.phase_el = angle.from_dms(90.)
        self.phase_az = angle.from_dms(0.)
        self.config = config
        self.timestamp = timestamp

    def set_visibilities(self, v, b):
        self.baselines = b
        self.v = v

    '''Rotated one, aimed at ra, decl

    Justification:

    Let
        s_1(t) = N(t) e^{j \omega t}

    t_g^{01) = t_g^1 - t_g^0 (arrival at a1 - arrival at a0)
    if t_g^{01} < 0, then s_0 arrives later than s_1 (defined by antennas.get_geo_delay_horizontal)
    therefore:

        s_0(t) = s_1(t - t_g^{01})

    v(0,1) = <s_0(t) s_1^{*}(t)>
                 = <s_1(t - t_g^{01}) s_1^{*}(t)>
                 = <N(t) e^{j \omega t} e^{-j \omega t_g^{01}} N^*(t) e^{-j \omega t}>
                 = e^{-j \omega t_g^{01}} <N(t) N^*(t)>

    So, after rotation

     <N(t) N^*(t)> = v(0,1) e^{+j \omega t_g^{01}}
    '''
    def rotate(self, sky_location):
        from tart.simulation import antennas
        stopped_vis = []
        omega = self.config.get_operating_frequency()*2.0*np.pi
        # Now we must do fringe stopping

        loc = self.config.get_loc()
        el, az = loc.equatorial_to_horizontal(self.timestamp, sky_location.ra, sky_location.dec)
        hsource =  HorizontalSource(r=9.0e99, azimuth=az, elevation=el)

        ant_pos = self.config.get_antenna_positions()
        for v, b in zip(self.v, self.baselines):
            a0 = antennas.Antenna(loc, ant_pos[b[0]])
            a1 = antennas.Antenna(loc, ant_pos[b[1]])

            tg = antennas.get_geo_delay_horizontal(a0, a1, hsource)
            # tg is t_a1 - t_a0
            # (negative if a1 is closer to source than a0)

            # print(b, omega*tg)
            v = v * np.exp(-1.0j * omega * tg)
            stopped_vis.append(v)

        self.phase_el = el
        self.phase_az = az
        self.v = stopped_vis

    def vis(self, i, j):
        if (j == i):
            raise "Baseline [%d,%d] is invalid" % (i,j)
        if (j < i): # The first index should be before the second
            return np.conjugate(self.vis(j,i))
        for k, b in enumerate(self.baselines):
            if (b == [i,j]):
                return self.v[k]
        raise "Baseline [%d,%d] is invalid" % (i,j)

    def get_closure_phase(self, i, j, k):
        return np.angle(self.vis(i,j)) + np.angle(self.vis(j,k)) - np.angle(self.vis(i,k))


    def toString(self):
        ret = ""
        for i,b in enumerate(self.baselines):
            ret += " V(%s)=%g, I%g" % (str(b), np.abs(self.v[i]), np.angle(self.v[i]))
        return ret



def Visibility_Lsq(vis1, vis2):
    """ Return least square based on the phases of 2 visibilities """
    if vis1.config.get_num_antenna() == vis2.config.get_num_antenna():
        difflist = []
        for v1, v2 in zip(vis1.v, vis2.v):
            diff = np.abs(np.angle(v1) - np.angle(v2))
            if diff > np.pi:
                diff = 2.*np.pi-diff
            difflist.append(diff)
        diffarr = np.array(difflist)
        return np.power(diffarr,2).sum()


def Visibility_From_Conf(config, timestamp, phase_el, phase_az):
    obs = observation.Observation(timestamp=timestamp, config=config)
    vis = Visibility(obs, phase_el, phase_az)
    return vis


def Visibility_Save_JSON(vis, filename):
    json_data = {}
    json_data['timestamp'] = vis.timestamp
    json_data['phase_el'] = vis.phase_el.to_degrees()
    json_data['phase_az'] = vis.phase_az.to_degrees()
    json_data['config'] = vis.config.Dict
    json_data['baselines'] = vis.baselines
    json_data['vis'] = vis.v
    with open(filename, 'w') as outfile:
        # The default=str handles datetime objects as strings
        json.dump(json_data, outfile, default=str, \
            sort_keys=True, indent=4, separators=(',', ': '))


'''
    Load and save lists of visibilities. Use filename extensions to deal with the deprecated .pkl files (or .vis files)
'''

def list_save(vis_list, filename):
    _, file_extension = os.path.splitext(filename)

    if ('.pkl' == file_extension):
        to_pkl(vis_list, filename)
    elif ('.vis' == file_extension):
        to_pkl(vis_list, filename)
    elif ('.hdf' == file_extension):
        to_pkl(vis_list, filename)
    else:
        raise RuntimeError("Unknown visibility file extension {}".format(file_extension))

def list_load(filename):
    _, file_extension = os.path.splitext(filename)

    if ('.pkl' == file_extension):
        vis_list = from_pkl(filename)
        err_count = 0
        ret = []
        for v in vis_list:
            if isinstance(v, tuple):
                err_count += 1
            else:
                ret.append(v)
        if err_count>0:
            print(('Warning. Visibility file: %s had %i visibilities missing' % (filename, err_count)))
        return ret
    elif ('.vis' == file_extension):
        vis_list = from_pkl(filename)
        return vis_list
    elif ('.hdf' == file_extension):
        vis_list = from_pkl(filename)
        return vis_list
    else:
        raise RuntimeError("Unknown file extension {}".format(file_extension))

'''
    Deal with Pickle files (Deprecated in favour of HDF5 files)
'''

def to_pkl(vis, filename):
    save_data = open(filename, 'wb')
    pickle.dump(vis, save_data, pickle.HIGHEST_PROTOCOL)
    save_data.close()

def from_pkl(filename):
    with open(filename, 'rb') as load_data:
        ret = pickle.load(load_data, encoding='latin1')
    return ret




'''
    Deal with HDF5 Files
'''

def to_hdf5(vis, filename):
    ''' Save the Observation object,
        to a portable HDF5 format
    '''
    with h5py.File(filename, "w") as h5f:
        dt = h5py.special_dtype(vlen=bytes)
        
        conf_dset = h5f.create_dataset('config', (1,), dtype=dt)
        conf_dset[0] = vis.config.to_json()

        ts_dset = h5f.create_dataset('timestamp', (1,), dtype=dt)
        ts_dset[0] = vis.timestamp.isoformat()

        h5f.create_dataset('baselines', data=vis.baselines)
        h5f.create_dataset('vis', data=vis.v)

        h5f.create_dataset('phase_elaz', data=[vis.phase_el.to_degrees(), vis.phase_az.to_degrees()])

def from_hdf5(filename):
    ''' Load the Visibility object,
        from a portable HDF5 format
    '''
    with h5py.File(filename, "r") as h5f:
        config_json = np.string_(h5f['config'][0])
        config = settings.from_json(config_json)
        timestamp = dateutil.parser.parse(h5f['timestamp'][0])
        
        hdf_vis = h5f['vis'][:]
        hdf_baselines = h5f['baselines'][:]
        hdf_phase_elaz = h5f['phase_elaz'][:]

    ret = Visibility(config=config, timestamp=timestamp)
    ret.set_visibilities(v=hdf_vis, b=hdf_baselines)
    ret.phase_el = hdf_phase_elaz[0]
    ret.phase_az = hdf_phase_elaz[1]
    
    return ret



import unittest
import numpy as np
import datetime

from tart.imaging.antenna_model import *
from tart.util.db import *

def create_empirical_antenna():
    sv = 1
    ant = EmpiricalAntenna(1)
    for e in np.arange(0., 89., 2.0):
        for a in np.arange(0., 359., 5.0):
            el = angle.from_dms(e)
            az = angle.from_dms(a)
            ant.add_measurement(el,az, e*(1+np.cos(np.pi/180.*a)), sv)
    for e in np.arange(0., -89., -2.0):
        for a in np.arange(0., 359., 5.0):
            el = angle.from_dms(e)
            az = angle.from_dms(a)
            ant.add_measurement(el,az, 0.0, sv)
    return ant

class TestAntennaModel(unittest.TestCase):
    '''Test our antenna models.'''


    def test_gps(self):
        '''Test the GPS patch antenna.'''
        ant = GpsPatchAntenna()
        self.assertEqual(ant.get_gain(angle.from_dms(120), angle.from_dms(120)), 0.0)
        self.assertEqual(ant.get_gain(angle.from_dms(80), angle.from_dms(120)), 1.0)
        self.assertEqual(ant.get_gain(angle.from_dms(-80), angle.from_dms(120)), 0.0)

    def test_ideal(self):
        '''Test the ideal hemispherical antenna'''
        ant = IdealHemisphericalAntenna()
        self.assertEqual(ant.get_gain(angle.from_dms(120), angle.from_dms(120)), 0.0)
        self.assertEqual(ant.get_gain(angle.from_dms(80), angle.from_dms(120)), 1.0)
        self.assertEqual(ant.get_gain(angle.from_dms(-80), angle.from_dms(120)), 0.0)

    #def test_empirical(self):
        #'''Test the empirical hemispherical antenna'''
        #ant = create_empirical_antenna()
        ##self.assertAlmostEqual(ant.get_gain(angle.from_dms(70), angle.from_dms(120)) / (70.*(1+np.cos(np.pi/180.*120.))), 1.0 , 1)
        ##self.assertAlmostEqual(ant.get_gain(angle.from_dms(45), angle.from_dms(45)) / (45.*(1+np.cos(np.pi/180.*45.))), 1.0, 1)
        ##self.assertAlmostEqual(ant.get_gain(angle.from_dms(-80), angle.from_dms(120)), 0.0, 1)
        #for e in np.arange(0., 89., 12.0):
            #for a in np.arange(0., 359., 25.0):
                #el = angle.from_dms(e)
                #az = angle.from_dms(a)
                #if e != 0:
                    #check = ant.get_gain(el, az) - e*(1.+np.cos(np.pi/180.*a))
                    #self.assertAlmostEqual(check, 0 , 1)

    def test_json_load(self):
        ant = create_empirical_antenna()
        ant.to_json('test.json')
        ant2 = EmpiricalAntenna.from_json('test.json')
        for e in np.arange(0, 89, 3.0):
            for a in np.arange(0, 359, 8.0):
                el = angle.from_dms(e)
                az = angle.from_dms(a)
                self.assertAlmostEqual(ant.get_gain(el, az), ant2.get_gain(el, az), 3)

    def test_db_load(self):
        ant = create_empirical_antenna()
        ant.to_db(utc_date=datetime.datetime.utcnow(), db_file='test.db')
        ant2 = EmpiricalAntenna.from_db(antenna_num=ant.antenna_num, db_file='test.db')
        for e in np.arange(0, 89, 3.0):
            for a in np.arange(0, 359, 8.0):
                el = angle.from_dms(e)
                az = angle.from_dms(a)
                self.assertAlmostEqual(ant.get_gain(el, az), ant2.get_gain(el, az), 3)





import numpy as np


from tart.simulation.skymodel import *
from tart.imaging import location

import unittest

class TestSkymodel(unittest.TestCase):
  """Testharness for Skymodel class"""
  def test_from_to_state_vector(self):
    """Generate Model from state vector. Then generate state vector from model and compare"""
    sm = Skymodel(5, location.Dunedin)
    state = sm.get_state_vector()
    sm2 = from_state_vector(state)
    state2 = sm2.get_state_vector()
    print(state)
    print(state2)

    # sv_init = np.array([15, 0, 90, -60, 2, 2, 0.5, 0.1])
    # model = from_state_vector(sv_init)
    # sv_final = model.get_state_vector()
    # for i, f in zip(sv_init, sv_final):
    #   self.assertAlmostEqual(i, f, 10)



# -*- coding: utf-8 -*-

import numpy as np
from tart.simulation import radio
from tart.simulation import antennas
from tart.simulation import simulation_source
from tart.operation import settings

                     
if __name__ == '__main__':
  import numpy as np

  rad = radio.Max2769B(sample_duration = 1e-5)
  sources = [simulation_source.SimulationSource(amplitude = 1.0, elevation = 30., azimuth = 0., sample_duration = rad.sample_duration)]

  print("Radio Sampling Rate %g" % rad.sampling_rate)
  
  config = settings.Settings('../../tools/operation/telescope_config.json')
  ants = [antennas.Antenna(config.get_loc(), pos) for pos in config.ant_positions]
  ant_sigs = antennas.antennas_new(ants, sources, rad.timebase)
  
  num_radio_samples = (len(rad.timebase) / rad.freq_mult) + 1  
  sampled_signals = np.zeros((config.get_num_antenna(), num_radio_samples))
  for i in range(0, config.get_num_antenna()):
   sampled_signals[i,:] = rad.sampled_signal(ant_sigs[i, :])

  print('The (truncated) array of radio signals is; ', sampled_signals[:, :10])
  print('Signal from each radio are in rows')



# Interpolator for sp3 orbit files
# Copyright 2013 (c) Tim Molteno. tim@elec.ac.nz

class Sp3Interpolator:
  
  def __init__(self, gpst, points):
    self.interps = points
    self.week = gpst.week()
  
  def lagrange_numerator(self, t_i, i, t):
    n = len(t_i)
    ret = 1.0
    for j in range(0,n):
       if (j != i):
         ret *= (t - t_i[j])
    return ret
  
  def lagrange_denominator(self, t_i, i):
    n = len(t_i)
    ret = 1.0
    for j in range(0, n):
      if (j != i):
        ret *= (t_i[i] - t_i[j])
    return ret
  
  def lagrange_interpolate(self, t_i, y_i, t):
    ret = 0.0
    n = len(t_i)
    for i in range(0, n):
      ret += y_i[i]*(self.lagrange_numerator(t_i, i, t) / self.lagrange_denominator(t_i, i))
    return ret
  
  def get_sv_position(self, gt, sv):
    sow = gt.sow()
    
    inter = self.interps[sv]
    
    dt0 = 999999.0
    i_close = 0
    n = len(inter)
    for j in range(0,n):
      i = inter[j]
      test = abs(i[0] - sow)
      if ((i[0] <= sow) and (test < dt0)):
        dt0  = test
        i_close = j
        
    t_points = []
    x_points = []
    y_points = []
    z_points = []

    order = 5
    i0 = max(0, i_close - order)
    i1 = min(i_close + order + 1, n-1)

    for i in range(i0,i1):
      t_points.append(inter[i][0])
      x_points.append(inter[i][1])
      y_points.append(inter[i][2])
      z_points.append(inter[i][3])
    
    xl = self.lagrange_interpolate(t_points, x_points, sow)*1000.0
    yl = self.lagrange_interpolate(t_points, y_points, sow)*1000.0
    zl = self.lagrange_interpolate(t_points, z_points, sow)*1000.0
    
    return [xl,yl,zl]



#calibration.py

import json

import numpy as np
from tart.util import constants

class CalibratedVisibility(object):
    def __init__(self, vis):
        self.vis = vis
        self.flagged_baselines = []
        self.phase_offset = np.zeros(self.get_config().get_num_antenna())
        self.tile_phase_offset = np.zeros(4)
        self.gain = np.ones(self.get_config().get_num_antenna())

    def set_config(self, config):
        self.vis.config = config

    def get_config(self):
        return self.vis.config

    def get_timestamp(self):
        return self.vis.timestamp

    def __get_vis(self,bl):
        return self.vis.v[self.vis.baselines.index(bl)]

    def flag_baseline(self,i,j):
        bl = [i,j]
        if bl not in self.flagged_baselines:
            #vis_idx = self.vis.baselines.index(bl)
            self.flagged_baselines.append(bl)
            # print('flagged', vis_idx, bl)

    def flag_antenna(self, i):
        for bl in self.vis.baselines:
          if i in bl:
            [a,b] = bl
            self.flag_baseline(a,b)

    def get_visibility(self,i,j):
        if (j < i):
            return np.conjugate(self.get_visibility(j, i))
        
        bl = [i,j]
        if bl in self.flagged_baselines:
            raise
        else:
            v = self.__get_vis(bl)
            return v * self.get_gain(i) * self.get_gain(j) * np.exp(-1j*(self.get_phase_offset(i)-self.get_phase_offset(j)))

    def get_all_visibility(self):
        bl = np.asarray(self.get_baselines())
        v_arr = np.asarray(self.get_unflagged_vis())
        ret = v_arr * self.get_gain(bl[:,0]) * self.get_gain(bl[:,1]) * np.exp(-1j*(self.get_phase_offset(bl[:,0])-self.get_phase_offset(bl[:,1])))
        return ret, bl

    def get_all_uvw(self):
        c = self.get_config()
        ant_p = np.asarray(c.get_antenna_positions())

        bls = np.asarray(self.get_baselines())
        bl_pos = ant_p[bls]
        uu_a, vv_a, ww_a = (bl_pos[:,0] - bl_pos[:,1]).T/constants.L1_WAVELENGTH
        return uu_a, vv_a, ww_a

    def get_unflagged_vis(self):
        return [vis for bl,vis in zip(self.vis.baselines, self.vis.v) if bl not in self.flagged_baselines]

    def get_baselines(self):
        return [bl for bl in self.vis.baselines if bl not in self.flagged_baselines]

    def get_baseline_lengths(self):
        pos = self.get_config().get_antenna_positions()
        #print(pos)
        ret = []
        for bln in self.get_baselines():
            [i,j] = bln
            ret.append([pos[j][0]-pos[i][0],pos[j][1]-pos[i][1]])
        return np.abs(ret)

    def set_flagged_baselines(self, flaged_list):
        self.flagged_baselines = flaged_list

    def flag_tile(self, tile_idx):
        for bln in self.get_baselines():
            [i,j] = bln
            if ((i//6) == tile_idx) or ((j//6) == tile_idx):
                self.flag_baseline(i,j)

    def leave_parallel_baselines(self,ew_threshold=10,ns_threshold=10):
        ant_positions = np.array(self.get_config().get_antenna_positions())
        for bln in self.get_baselines():
          [i,j] = bln
          diff = ant_positions[i]-ant_positions[j]
          if np.abs(diff[0])>ew_threshold or np.abs(diff[1])>ns_threshold:
              self.flag_baseline(i,j)

    def leave_intra_tile_baselines(self):
        ant_positions = np.array(self.get_config().get_antenna_positions())
        for bln in self.get_baselines():
          [i,j] = bln
          if ((i//6) != (j//6)):
            self.flag_baseline(i,j)

    def set_phase_offset(self, i, val):
        self.phase_offset[i] = np.asarray(val) % (2*np.pi)

    def get_phase_offset(self, i):
        return self.phase_offset[i]+self.get_tile_phase_offset_for_antenna(i)

    def set_tile_phase_offset(self,i,val):
        self.tile_phase_offset[i] = val

    def get_tile_offset(self, i):
        return self.tile_phase_offset[i]

    def get_tile_phase_offset_for_antenna(self, ant_idx):
        return self.tile_phase_offset[np.asarray(ant_idx)//6]

    def set_gain(self, i, val):
        self.gain[i] = val

    def get_gain(self, i):
        return self.gain[i]

    def to_json(self, filename='gain_calibration.json'):
        calib_dict = {}
        calib_dict['gain'] = [int(1e5*self.get_gain(i))/float(1e5) for i in range(len(self.gain))]
        calib_dict['phase_offset'] = [int(1e5*self.get_phase_offset(i))/float(1e5) for i in range(len(self.gain))]
        calib_dict['flagged_baselines'] = self.flagged_baselines
        with open(filename, 'w') as handle:
          json.dump(calib_dict, handle, sort_keys = True, indent = 4,
               ensure_ascii = False)
        json_str = json.dumps(calib_dict)
        return json_str

def from_dict(vis, calib_dict):
    ret = CalibratedVisibility(vis)
    ret.set_phase_offset(list(range(ret.get_config().get_num_antenna())),np.array(calib_dict['phase_offset']))
    ret.set_gain(list(range(ret.get_config().get_num_antenna())),np.array(calib_dict['gain']))
    ret.set_flagged_baselines(calib_dict['flagged_baselines'])
    return ret

def from_JSON(vis, json_str):
    calib_dict = json.loads(json_str)
    return from_dict(vis, calib_dict)

def from_JSON_file(vis, filename='gain_calibration.json'):
    with open(filename) as data_file:
      calib_dict = json.load(data_file)
    return from_dict(vis, calib_dict)

import argparse

from tart.imaging import visibility



if __name__ == '__main__':
  parser = argparse.ArgumentParser(description='Generate visibilities for a skymodel in parallel.')
  parser.add_argument('--vis', default="/freenas2/tart/visibility/vis_2015_08_10.vis",required=False, help="")
  args = parser.parse_args()

  vis = visibility.list_load(args.vis) #[:20]

  cal_vis = CalibratedVisibility(vis[0])
  #print(cal_vis.get_visibility(0,1))
  #print(cal_vis.get_baselines())
  #print(len(cal_vis.get_baselines()))

  cal_vis.flag_baseline(0,1)

  #print(cal_vis.get_baselines())
  #print(len(cal_vis.get_baselines()))


  def gen_tile_vis_list(tile_no,vis):
    for v in vis:
      cal_vis_list = []
      cv = calibration.CalibratedVisibility(v)
      for i in [0,1,2,3]:
        if i!=tile_no:
          cv.flag_tile(i)

    # cv.flag_tile(1)
    # cv.flag_tile(2)
    # cv.flag_tile(3)
    # cv.leave_parallel_baselines(ns_threshold=0.1)
    # cv.flag_baseline(0,5)
    cal_vis_list.append(cv)
    return cal_vis_list

  # t0_vis_list = gen_tile_vis_list(0,vis)
  # t1_vis_list = gen_tile_vis_list(1,vis)
  # t2_vis_list = gen_tile_vis_list(2,vis)
  # t3_vis_list = gen_tile_vis_list(3,vis)
  # fig, axarr = plt.subplots(2,2,sharex=True,sharey=True)
  # t_list = [t0_vis_list, t1_vis_list, t2_vis_list, t3_vis_list]
  # ift_arr = []
  # for i, ax in enumerate(axarr.ravel()):
  #   syn_tile = synthesis.Synthesis_Imaging(t_list[i])
  #   ift = syn_tile.get_image(nw=50,num_bin=2**9,ex_ax=ax)
  #   ift_arr.append(ift)


  # ift_arr = np.array(ift_arr)
  # print(ift_arr)
  # plt.figure()
  # plt.imshow(np.prod(np.abs(ift_arr),axis=0))


  # print(cal_vis.get_visibility(0,1))



# -*- coding: utf-8 -*-

# This module calculates the length of sample required from each
# astronomical radio noise source given parameters including
# the phase delays required

# An example to aid understanding;
# -3 -2 -1 0 1 2 3 4 5 ... E E+1 E+2 E+3	A sampled radio source	
#     ^						0th pos phase leads 2
#              ^				Oth pos phase lags 2

# Max lag represents the position of the zero value
# Phase Lead is positive number, Phase Lag is a negative number
import numpy as np

def sample_size(ref_freq, freq_mult, sample_duration, sample_delay):
  
  # Calculate the required sample size excluding offsets
  base_size = int(ref_freq * freq_mult * sample_duration)
  
  # Calculate max and min offsets for each signal
  extrema = np.zeros((2, sample_delay.shape[0])) # To hold extreme offsets
  extrema[0] = sample_delay.max(axis=1) # Max values
  extrema[1] = sample_delay.min(axis=1) # Min values
  
  # Calculate the sample size given the offsets
  sizes = np.ones(extrema.shape[1]) # blank array to hold sample sizes
  for sig in range(0, extrema.shape[1]):
    sizes[sig] = base_size + extrema[0, sig] - extrema[1, sig] # Row 1 -ve vals
    
  max_size = sizes.max(0)
    
  # The maximum values indicate 
  zero_off = extrema[0]

  return zero_off, max_size, base_size
  
if __name__ == '__main__':
  
  ref_freq = 16.368e6
  freq_mult = 256
  sample_duration = 1e-3
  
  sample_delay = np.array([[-49, -39, -29, -19, -9, 0, 9, 19, 29, 39, 49],
  [-34, -27, -20, -13, -6, 0, 6, 13, 20, 27, 34],
  [-18, -14, -10, -7, -3, 0, 3, 7, 10, 14, 18],
  [  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
  [ 18, 14, 10, 7, 3, 0, -3, -7, -10, -14, -18],
  [ 34, 27, 20, 13, 6, 0, -6, -13, -20, -27, -34],
  [ 49, 39, 29, 19, 9, 0, -9, -19, -29, -39, -49]])
 
  zero_off, max_size, base_size = sample_size(ref_freq, freq_mult, sample_duration, sample_delay)
  print('Jump this many places from left to find phase zero position for each sample')
  print(zero_off)
  print('The base size is ', base_size)
  print('Length of sample required to accommodate all phase changes is ', max_size)



# -*- coding: utf-8 -*-
import numpy as np
from pylab import plot, show, title, xlabel, ylabel, subplot, legend
from scipy import fft, arange

def plotSpectrum(y,Fs, c='red', label='powerspectrum'):
    """
    Plots a Single-Sided Amplitude Spectrum of y(t)
    """
    n = len(y) # length of the signal
    k = arange(n)
    T = n/Fs
    frq = k/T # two sides frequency range
    frq = frq[list(range(n/2))] # one side frequency range

    Y = fft(y)/n # fft computing and normalization
    Y = Y[list(range(n/2))]

    plot(frq,abs(Y), c=c, label=label) # plotting the spectrum
    xlabel('Freq (Hz)')
    ylabel('|Y(freq)|')
    return (frq,np.array(abs(Y)))

if __name__ == '__main__':
    Fs = 150.0;  # sampling rate
    Ts = 1.0/Fs; # sampling interval
    t = np.arange(0,1,Ts) # time vector

    ff = 5;   # frequency of the signal
    y = np.sin(2*np.pi*ff*t)

    subplot(2,1,1)
    plot(t,y)
    xlabel('Time')
    ylabel('Amplitude')
    subplot(2,1,2)
    plotSpectrum(y,Fs)
    show()



"""Copyright (C) Max Scheel 2013. All rights reserved"""

from tart.util import angle

#from tart.imaging import sun
from tart.imaging import radio_source
from tart.imaging import gps_satellite
from tart.imaging import location

from tart.simulation import simulation_source

import numpy as np
import healpy as hp
import matplotlib.pyplot as plt

import requests
import json
import datetime


def get_L1_srcs(ts=None):
  '''
    In order to reduce the number of requests to the catalog it is adviseable to establish a local nginx proxy:
    apt-get install nginx
    vim /etc/nginx/sites-available/default

    proxy_cache_path /data/nginx/cache levels=1:2 keys_zone=my_cache:10m max_size=10g;
    server {
    ...
          location /catalog/ {
                proxy_cache my_cache;
                proxy_cache_methods GET HEAD POST;
                proxy_ignore_headers Cache-Control;
                proxy_cache_valid any 30m;
                proxy_pass https://tart.elec.ac.nz/catalog/;
          }
  '''
  # catalog_server = "https://tart.elec.ac.nz/catalog"

  catalog_server = "http://localhost/catalog"
  if ts is None:
    ts = datetime.datetime.utcnow()
    r = requests.get('{}/catalog?lat=-45.851820&lon=170.545448&date={}'.format(catalog_server,ts.isoformat()))
  else:
    r = requests.get('{}/catalog?lat=-45.851820&lon=170.545448&date={}'.format(catalog_server,ts.isoformat()))
  return np.array(json.loads(r.text))



class Skymodel(object):
  """Ensemble of sources and their visibilities"""
  def __init__(self, n_sources, location, sun_str=2.e5, sat_str=5.01e6, gps=True, l1_catalog=False, l1_elevation_threshold = 30, l1_allowed=[], thesun=False, known_cosmic=True):
    self.n_sources = n_sources
    self.source_list = []
    self.known_objects = []
    self.gps_ants = []
    self.location = location

    self.l1_catalog = l1_catalog
    self.l1_elevation_threshold = l1_elevation_threshold
    self.l1_allowed = l1_allowed

    self.sun_str=sun_str
    self.sat_str=sat_str
    self.gps = gps
    self.thesun = thesun
    self.known_cosmic= known_cosmic

    self.el_threshold = 0

    if self.thesun:
      self.add_src(sun.Sun(jy=sun_str))
    if self.gps:
      for i in range(32):
        sv = gps_satellite.GpsSatellite(i+1, location=self.location,jy=sat_str)
        self.add_src(sv)
    #if self.known_cosmic:
    #  for src in radio_source.BrightSources:
    #    self.add_src(src)
    #for _ in range(self.n_sources):
      #ra = angle.from_dms(np.random.uniform(0., 360.))
      #dec = angle.asin(np.random.uniform(-1., 1))
      #cs = radio_source.CosmicSource(ra, dec)
    #  self.add_src(cs)

  def gen_beam(self, utc_date_init, utc_date_obs, config, radio,  az_deg = 20., el_deg = 80.):
    ''' Generate point source with constant at RA and DEC according to given az and el at time utc_date_init'''
    sources = []
    ra, dec = location.get_loc(config).horizontal_to_equatorial(utc_date_init, angle.from_dms(el_deg), angle.from_dms(az_deg))
    src = radio_source.CosmicSource(ra, dec)
    el, az = src.to_horizontal(location.get_loc(config), utc_date_obs)
    sources.append(simulation_source.SimulationSource(amplitude = 1., azimuth = az, elevation = el, \
      sample_duration = radio.n_samples/radio.ref_freq))
    return sources

  def get_cum_src_flux(self, utc_date):
    '''Return cumulative flux as list over all sources'''
    cumulativ_src_flux = []
    for k in range(len(self.known_objects)+1):
      cumulativ_src_flux.append(np.array([src.jansky(utc_date) for src in self.known_objects[:k] ]).sum())
    return cumulativ_src_flux

  def get_int_src_flux(self, utc_date):
    '''Return cumulative flux'''
    return self.get_cum_src_flux(utc_date)[-1]

  def gen_photons_per_src(self, utc_date, radio, config, n_samp=1):
    ''' Generate n_samp photons per source'''
    sources = []
    #for src in self.known_objects:
    if self.l1_catalog:
      for src in get_L1_srcs(utc_date):
        if (src['el']>self.l1_elevation_threshold):
          l = len(self.l1_allowed)
          if (l==0) or (src['name'] in self.l1_allowed['names']):
            if (l!=0):
              s_str = self.l1_allowed['strength_log10'][self.l1_allowed['names'].index(src['name'])]
              #print(s_str, type(s_str))
              amplitude = np.power(10,s_str)
            else:
              amplitude = 1.
            print(src['name'], src['az'], src['el'])
            sources.append(simulation_source.SimulationSource(\
            r = src['r'],
            amplitude = amplitude, \
            azimuth = angle.from_dms(src['az']), elevation = angle.from_dms(src['el']), sample_duration = radio.n_samples/radio.ref_freq))
            #print('here')

    for src in self.known_objects:
      print('extra',src)
      #for src in self.get_src_objects(location.get_loc(config), utc_date):
      ra, declination = src.radec(utc_date)
      dx, dy = np.random.multivariate_normal([0., 0.], np.identity(2)*np.power(src.width, 2.), n_samp).T
      for j in range(n_samp):
        el, az = location.get_loc(config).equatorial_to_horizontal(utc_date, \
          ra + angle.from_dms(dx[j]), declination + angle.from_dms(dy[j]))
        sources.append(simulation_source.SimulationSource(\
          r = src.r,
          amplitude = src.jansky(utc_date)/self.get_int_src_flux(utc_date)*1./n_samp, \
          azimuth = az, elevation = el, sample_duration = radio.n_samples/radio.ref_freq))
    print(len(sources))
    return sources


  def gen_n_photons(self, config, utc_date, radio, n=10):
    ''' Generate a total of n photons. Sources with more jansky will contribute more photons'''
    cumulativ_src_flux = self.get_cum_src_flux(utc_date)
    int_src_flux = cumulativ_src_flux[-1]
    rel_noise_flux = 0.0
    tot_flux = int_src_flux * (1. + rel_noise_flux)
    src_identifier = np.random.uniform(0., tot_flux, n)
    hi, _ = np.histogram(src_identifier, bins=cumulativ_src_flux)
    ret = []
    for src_num, count in enumerate(hi):
      src = self.known_objects[src_num]
      # Assume the sky is flat.
      # this will also cause problems at problems at boundaries of dec.
      dx, dy = np.random.multivariate_normal([0., 0.], np.identity(2)*np.power(src.width, 2.), count).T
      ra, declination = src.radec(utc_date)
      for j in range(count):
        el, az = location.get_loc(config).equatorial_to_horizontal(utc_date, ra\
          + angle.from_dms(dx[j]), declination + angle.from_dms(dy[j]))
        ret.append(simulation_source.SimulationSource(amplitude = 1./n, \
            azimuth = az, elevation = el, sample_duration = radio.n_samples/radio.ref_freq))
    return ret

  def set_el_threshold(self, threshold=0):
    self.el_threshold = threshold

  def get_src_objects(self, location, utc_date, threshold=None, debug=False):
    '''optional argument threshold defines elevation threshold in deg. 0 = horizon. 90 = zenith'''
    if threshold is None:
      threshold = self.el_threshold
    ret = []
    for src in self.known_objects:
      try:
        el, az = src.to_horizontal(location=location, utc_date=utc_date)
        if (el.to_degrees() > threshold):
          ret.append(src)
      except:
        print('No position for', src)
        self.del_src(src)
    return ret

  def add_src(self, src):
    '''Add source to known_objects'''
    self.known_objects.append(src)

  def del_src(self, src):
    '''Remove source from known_objects'''
    self.known_objects.remove(src)

  def get_state_vector(self):
    '''Return state vector'''
    state_vector = np.zeros(self.n_sources*4)
    for i, source in enumerate(self.source_list):
      state_vector[i+(0*self.n_sources)] = source.skyloc.ra.to_degrees()
      if source.skyloc.dec.to_degrees() > 90.:
        state_vector[i+(1*self.n_sources)] = source.skyloc.dec.to_degrees() - 360.
      else:
        state_vector[i+(1*self.n_sources)] = source.skyloc.dec.to_degrees()
      state_vector[i+(2*self.n_sources)] = source.flux
      state_vector[i+(3*self.n_sources)] = source.width
    return [self.sun_str, self.sat_str, self.gps, self.thesun, self.known_cosmic, self.location, state_vector]

  def get_src_positions(self, location, utc_date):
    ''' Save way of getting lists of el and az for all known_objects'''
    l_el = []
    l_az = []
    l_name = []
    for src in self.known_objects:
      try:
        el, az = src.to_horizontal(location, utc_date)
        if (el.to_rad()>0):
          l_el.append(el.to_rad())
          l_az.append(az.to_rad())
          l_name.append(str(src))
          # l_name.append('%s %2.2f %1.1e' % (str(src), el.to_degrees(), src.jansky(utc_date)))
      except:
        print('no position', src)
    return l_el, l_az, l_name

  def true_image(self, location, utc_date, nside):
    '''Plot current sky.'''

    l_el, l_az, l_name = self.get_src_positions(location, utc_date)

    m = np.zeros(hp.nside2npix(nside))*hp.UNSEEN
    th = -np.array(l_el) + np.pi/2.
    l_phi = -np.array(l_az)
    pix = hp.pixelfunc.ang2pix(nside, th, l_phi)
    m[pix] = 1 #[src.jansky(utc_date) for src in self.known_objects]
    m = hp.sphtfunc.smoothing(m,fwhm=3.*np.pi/180.)
    # plt.figure()
    # hp.visufunc.orthview(m, title="", rot=(0,90,0), )
    return m


  def true_image_overlay(self, location, utc_date):
    '''Plot current sky.
    Theta is colatitude and measured from North pole. 0 .. pi
     (straight up) |  el:  pi/2   | theta 0
     (horizon)     |  el:  0      | theta pi/2
    Th = pi/2 - el

    flip : {'astro', 'geo''}, optional
    Defines the convention of projection : 'astro'' (default, east towards left, west towards right) or 'geo' (east towards right, west towards left)
    '''
    l_el, l_az, l_name = self.get_src_positions(location, utc_date)

    th = np.pi/2. - np.array(l_el)
    l_phi = -np.array(l_az)
    #_ = [hp.projtext(i, j, n, rot=(0,90,0), color='black', weight='light', ha='left', va='center') for i, j, n in zip(th, l_phi, l_name)]
    _ = [hp.projtext(i, j, n, rot=(0,90,0), color='gray', alpha=0.8, weight='bold', ha='center', va='bottom') for i, j, n in zip(th, l_phi, l_name)]
    _ = [hp.projscatter(i, j, rot=(0,90,0), color='black', alpha=1.0, s=25) for i, j, n in zip(th, l_phi, l_name)]
    _ = [hp.projscatter(i, j, rot=(0,90,0), color='white', alpha=1.0, s=10) for i, j, n in zip(th, l_phi, l_name)]

    hp.projtext(np.pi/2, 0., 'N', rot=(0,90,0), va='top', ha='center')
    hp.projtext(np.pi/2, -np.pi/2., 'E', rot=(0,90,0), va='center')
    hp.projtext(np.pi/2, -np.pi, 'S', rot=(0,90,0), ha='center')
    hp.projtext(np.pi/2, -np.pi*3./2., 'W', rot=(0,90,0), ha='right', va='center')


  def true_FFT_overlay(self, utc_date, nw, num_bin):
    '''Plot current sky.
    Theta is colatitude and measured from North pole. 0 .. pi
     (straight up) |  el:  pi/2   | theta 0
     (horizon)     |  el:  0      | theta pi/2
    Th = pi/2 - el

    flip : {'astro', 'geo''}, optional
    Defines the convention of projection : 'astro'' (default, east towards left, west towards right) or 'geo' (east towards right, west towards left)
    '''
    from tart.imaging.synthesis import get_max_ang

    max_ang = get_max_ang(nw, num_bin)
    l_el, l_az, l_name = self.get_src_positions(self.location, utc_date)
    l_el, l_az = np.array(l_el), np.array(l_az)

    r = (90.-np.degrees(l_el))
    x = r * np.sin(l_az)
    y = r * np.cos(l_az)
    plt.scatter(x, y, color='white', s=50)

def from_state_vector(state):
  '''Generate skymodel from state vector'''
  sun_str, sat_str, gps, thesun, known_cosmic, location, state_vector = state
  psky = Skymodel(0, location=location, sun_str=sun_str, \
    sat_str=sat_str, gps=gps, known_cosmic=known_cosmic)
  psky.n_sources = len(state_vector)/4
  psky.source_list = []

  for i in range(psky.n_sources):
    ra = angle.from_dms(state_vector[i+(0*psky.n_sources)])
    dec = angle.from_dms(state_vector[i+(1*psky.n_sources)])
    gs = radio_source.CosmicSource(ra, dec, jy=state_vector[i+(2*psky.n_sources)], width=state_vector[i+(3*psky.n_sources)])
    psky.source_list.append(gs)
  return psky



"""Copyright (C) Max Scheel 2016. All rights reserved"""
import numpy as np
from multiprocessing import Pool
from tart.simulation import antennas
import time
try:
   import cPickle as pickle
except:
   import pickle


def get_vis_parallel(sky, cor, rad, ants, ant_models, config, time, mode='simp'):
  p = Pool()
  resultList = []
  ret_vis_list = []
  try:
    if type(sky) in [np.ndarray, list]:
      for s in sky:
        resultList.append(p.apply_async(get_vis, (s, cor, rad, ants, ant_models, config, time, mode)))
    if type(time) in [np.ndarray, list]:
      for t in time:
        resultList.append(p.apply_async(get_vis, (sky, cor, rad, ants, ant_models, config, t, mode)))
    p.close
    p.join
    for thread in resultList:
      vis = thread.get()
      if (vis != None):
        ret_vis_list.append(vis)
    p.terminate()
  except KeyboardInterrupt:
    print('control-c pressed')
    p.terminate()
  return ret_vis_list

def get_vis(sky, cor, rad, ants, ant_models, config, timestamp, mode='simp', seed=None,):
  np.random.seed(seed=seed)
  sources = sky.gen_photons_per_src(timestamp, radio=rad, config=config, n_samp=1)
  # sources = sky.gen_n_photons(config, timestamp, radio=rad, n=10)
  # print('debug: total flux',  np.array([src.jansky(timestamp) for src in self.known_objects]).sum())
  # print('debug: total amplitude', np.array([src.amplitude for src in sources]).sum())

  if mode == 'full':
    timebase = np.arange(0, rad.sample_duration, 1.0/rad.sampling_rate)
    ant_sigs_full = antennas.antennas_signal(ants, ant_models, sources, timebase)
    obs = rad.get_full_obs(ant_sigs_full, timestamp, config, timebase)
    v = cor.correlate_roll(obs)

  elif mode == 'mini':
    v = antennas.antennas_simp_vis(ants, ant_models, sources, timestamp, config, rad.noise_level)
  else:
    ant_sigs_simp = antennas.antennas_simplified_signal(ants, ant_models, sources, rad.baseband_timebase, rad.int_freq, seed=seed)
    obs = rad.get_simplified_obs(ant_sigs_simp, timestamp, config = config, seed=seed)
    v = cor.correlate_roll(obs)
  return v


def get_vis_parallel_segmented(sky, cor, rad, ants, ant_models, config, time, mode='simp', segment_dir='temp/'):
  vis_per_segment = 1000
  num_segments = int(np.ceil(len(sky)/float(vis_per_segment)))

  for i in range(num_segments):
    print('computing segment %i of %i' %(i, num_segments))
    start_idx = i*vis_per_segment
    end_idx = (i+1)*vis_per_segment
    vis_seg = get_vis_parallel(sky[start_idx:end_idx], cor, rad, ants, ant_models, config, time, mode='simp')
    pickle.dump( vis_seg, open( segment_dir+'segment_%010i.pkl' %i, "wb" ) )

  #//print('now combining segments.')
  ret = []
  #print(num_segments)
  #for i in range(num_segments):
    #print(i)
    #vis_seg = pickle.load( open( segment_dir+'segment_%010i.pkl' %i, "rb" ) )
    #ret += vis_seg
  return ret



# Copyright (C) Tim Molteno 2008-2019. All rights reserved

import datetime
import math

from tart.util import utc

class GpsTime(object):
    SEC_PER_DAY = 86400.0

    def __init__(self, y, m, d, h, minutes, seconds):
        if (y < 1980):
                raise Exception("Year #{y} must >= 1980")

        in_time = utc.utc_datetime(year=y, month=m, day=d, hour=h, minute=minutes, second=0)
        gps_t = utc.utc_datetime(year=1980, month=1, day=6, hour=0, minute=0, second=0)

        self.m_day_of_year = in_time.timetuple().tm_yday

        gps_sec = ( in_time - gps_t ).total_seconds() + seconds + self.utc_offset(in_time)
        gps_days = int(math.floor(gps_sec / GpsTime.SEC_PER_DAY))
        gps_weeks = gps_days / 7.0

        self.m_week = int(math.floor(gps_weeks))
        self.sec_of_week = gps_sec - self.m_week*7.0*GpsTime.SEC_PER_DAY


    @classmethod
    def from_time(self, utc_date):
        return GpsTime(utc_date.year, utc_date.month, utc_date.day, utc_date.hour, utc_date.minute, utc_date.second)

    # I don't think that the UTC offset works correctly.
    @classmethod
    def utc_offset(self, t):
        if (t >= utc.utc_datetime(2017, 1, 1)):
            return 18
        if (t >= utc.utc_datetime(2015, 7, 1)):
            return 17
        if (t >= utc.utc_datetime(2012, 7, 1)):
            return 16
        if (t >= utc.utc_datetime(2009, 1, 1)):
            return 15
        if (t >= utc.utc_datetime(2006, 1, 1)):
            return 14
        if (t >= utc.utc_datetime(1999, 1, 1)):
            return 13
        if (t >= utc.utc_datetime(1997, 7, 1)):
            return 12
        if (t >= utc.utc_datetime(1996, 1, 1)):
            return 11
        if (t >= utc.utc_datetime(1994, 7, 1)):
            return 10
        if (t >= utc.utc_datetime(1993, 7, 1)):
            return 9
        if (t >= utc.utc_datetime(1992, 7, 1)):
            return 8
        if (t >= utc.utc_datetime(1991, 1, 1)):
            return 7
        if (t >= utc.utc_datetime(1990, 1, 1)):
            return 6
        if (t >= utc.utc_datetime(1988, 1, 1)):
            return 5
        if (t >= utc.utc_datetime(1985, 7, 1)):
            return 4
        if (t >= utc.utc_datetime(1983, 7, 1)):
            return 3
        if (t >= utc.utc_datetime(1982, 7, 1)):
            return 2
        if (t >= utc.utc_datetime(1981, 7, 1)):
            return 1
        if (t >= utc.utc_datetime(1980, 1, 1)):
            return 0
        return False


    def __repr__(self):
        return "W:%d_SOW:%d (doy %d)" % (self.m_week, self.sec_of_week, self.m_day_of_year)


    def diff(self,b):
        w = (self.m_week - b.week()) * 7 * GpsTime.SEC_PER_DAY
        s = (self.sec_of_week - b.sow())
        return w + s


    def week(self):
        return self.m_week


    def sow(self):
        return self.sec_of_week


    def day_of_year(self):
        return self.m_day_of_year


    def day_of_week(self):
        dow = self.sec_of_week / GpsTime.SEC_PER_DAY;
        return int(math.floor(dow))



import unittest

import numpy as np
import datetime
import matplotlib.pyplot as plt


from tart.imaging import antenna_model
from tart.operation import settings
from tart.operation import observation
from tart.simulation import simulation_source
from tart.simulation import radio
from tart.simulation import antennas
from tart.util import angle
from tart.imaging import correlator

TEST_CONFIG='./tart/test/test_telescope_config.json'
TEST_ANTENNA_POSITIONS='./tart/test/test_calibrated_antenna_positions.json'

def relative_diff(l1, l2):
  return (l1-l2)/np.sqrt(np.power(l1, 2) + np.power(l2, 2))

class TestSimulatedVisibility(unittest.TestCase):

  def setUp(self):
    self.utc_date = datetime.datetime.utcnow()
    self.config = settings.from_file(TEST_CONFIG)
    self.config.load_antenna_positions(cal_ant_positions_file=TEST_ANTENNA_POSITIONS)

    self.rad = radio.Max2769B(noise_level=[0. for _ in range(self.config.get_num_antenna())])
    self.sources = [simulation_source.SimulationSource(r=1e9,amplitude = 1.0, azimuth = angle.from_dms(2.0), elevation = angle.from_dms(50.0), sample_duration = self.rad.sample_duration)]
    self.ants = [antennas.Antenna(self.config.get_loc(), pos) for pos in self.config.get_antenna_positions()]
    self.ant_models = [antenna_model.GpsPatchAntenna() for i in range(self.config.get_num_antenna())]
    self.cor = correlator.Correlator()
    self.timebase = np.arange(0, self.rad.sample_duration, 1.0/self.rad.sampling_rate)

    ant_sigs_full = antennas.antennas_signal(self.ants, self.ant_models, self.sources, self.timebase)
    obs = self.rad.get_full_obs(ant_sigs_full, self.utc_date, self.config, self.timebase)
    vis = self.cor.correlate(obs)
    self.full_vis = np.array(vis.v)

    ant_sigs_simp = antennas.antennas_simplified_signal(self.ants, self.ant_models, self.sources, self.rad.baseband_timebase, self.rad.int_freq)
    obs = self.rad.get_simplified_obs(ant_sigs_simp, self.utc_date, config = self.config)
    vis2 = self.cor.correlate(obs)
    self.sim_vis = np.array(vis2.v)

    plt.figure(figsize=(4,4))
    plt.scatter(self.full_vis.real,self.full_vis.imag, label='full')
    plt.scatter(self.sim_vis.real,self.sim_vis.imag, color='red', label='simpl')
    plt.legend()
    plt.show()
  # def test_full_vis(self):
  #   self.assertGreater(self.full_vis.max(), 0.8)
  #   self.assertLess(self.full_vis.min(), -0.8)

  # def test_simp_vis(self):
  #   self.assertGreater(self.sim_vis.max(), 0.8)
  #   self.assertLess(self.sim_vis.min(), -0.8)

  def test_compare_vis(self):
    d = self.full_vis - self.sim_vis
    print(d)
    print('relative diffs', relative_diff(self.full_vis, self.sim_vis))
    self.assertLess(abs(d.max()), 0.3)
    self.assertLess(d.std(), 0.1)

class TestTelescope(unittest.TestCase):
  def setUp(self):

    # rad = radio.Max2769B(noise_level=0.)
    utc_date = datetime.datetime.utcnow()
    self.config = settings.from_file(TEST_CONFIG)
    self.config.load_antenna_positions(cal_ant_positions_file=TEST_ANTENNA_POSITIONS)

    rad = radio.Max2769B(noise_level=[0. for _ in range(self.config.get_num_antenna())])
    sources = [simulation_source.SimulationSource(r=1e9, amplitude = 1.0, azimuth = angle.from_dms(2.0), elevation = angle.from_dms(50.0), sample_duration = rad.sample_duration)]
    ant_models = [antenna_model.GpsPatchAntenna() for i in range(self.config.get_num_antenna())]
    ants = [antennas.Antenna(self.config.get_loc(), pos) for pos in self.config.get_antenna_positions()]
    self.timebase = np.arange(0, rad.sample_duration, 1.0/rad.sampling_rate)

    ant_sigs = antennas.antennas_signal(ants, ant_models, sources, self.timebase)
    self.obs = rad.get_full_obs(ant_sigs, utc_date, self.config, self.timebase)

  def test_data_limits(self):
    for i in range(1, self.obs.config.get_num_antenna()):
      data = self.obs.get_antenna(i)

      self.assertGreater(data.max(), 0.99)
      self.assertLess(data.min(), -0.99)

      self.assertGreater(data.mean(), -0.1)
      self.assertLess(data.mean(), 0.1)

  def test_settings(self):
    self.assertEqual(self.config.get_num_antenna(), 24)

  def test_simulated_observation(self):
    fname = 'test_observation.pkl'
    self.obs.save(fname)
    nobs = observation.Observation_Load(fname)
    self.assertEqual(self.obs.timestamp, nobs.timestamp)
    self.assertTrue((self.obs.get_antenna(1) == nobs.get_antenna(1)).all())



import datetime
import unittest

from astropy import time 
from astropy.time import Time 

from tart.imaging.tart_util import *
from tart.util.utc import now

class TestTartUtil(unittest.TestCase):

    def test_astropy(self):
        utcd = now()
        obstime = time.Time(utcd, scale='utc')

        jd = JulianDay(utcd)
        
        self.assertAlmostEqual(obstime.jd, jd, 4)





import math

class Util(object):
    PI  = 3.1415926535897932
    PI2 = 6.2831853071795864
    C   = 2.99793e8

    @classmethod
    def check_t(self, t):
        half_week = 302400.0
        tt = t

        if (t > half_week):
             tt = t-2*half_week
        if (t < -half_week):
            tt = t+2*half_week

        return tt

    @classmethod
    def idiv(self, x, y):
        ''' Integer Division '''
        return int(math.floor(x / y))


    @classmethod
    def rem(self, x, y):
        ''' Remainder '''
        return x - y * Util.idiv(x,y)


    #! \brief remainder = numerator - quotient * denominator
    #
    @classmethod
    def mod(self, x, y):
        ret = x - y*Util.idiv(x,y)
        return ret;


    @classmethod
    def mod_int(self, x, y):
        ret = x%y
        if (x < 0):
            ret = y - (-x)%y
        return ret


    @classmethod
    def rad2deg(self, x):
        return x*180.0 / Util.PI


    @classmethod
    def rem2pi(self, x):
        return Util.rem(x, Util.PI2)




import unittest

import numpy as np
import datetime
import matplotlib.pyplot as plt

from tart.operation import settings
from tart.simulation import simulation_source
from tart.simulation import antennas
from tart.simulation import spectrum
from tart.imaging import antenna_model
from tart.util import angle

from tart.simulation.radio import *



class TestMax2769B(unittest.TestCase):
  def setUp(self):
    self.config = settings.from_file('./tart/test/test_telescope_config.json')
    self.config.load_antenna_positions(cal_ant_positions_file='./tart/test/test_calibrated_antenna_positions.json')
    # noiselvls =  0.1.*np.ones(config.get_num_antenna())
    num_ant = self.config.get_num_antenna()
    noiselvls =  0. * np.ones(num_ant)
    self.rad = Max2769B(noise_level = noiselvls)
    self.sources = [simulation_source.SimulationSource(r=1e9,amplitude = 1.0, azimuth = angle.from_dms(0.), elevation = angle.from_dms(90.), sample_duration = self.rad.sample_duration)]
    self.ants = [antennas.Antenna(self.config.get_loc(), pos) for pos in self.config.get_antenna_positions()]
    self.ant_models = [antenna_model.GpsPatchAntenna() for i in range(num_ant)]
    self.utc_date = datetime.datetime.utcnow()

  def test_get_obs(self):

    plt.figure()
    timebase = np.arange(0, self.rad.sample_duration, 1.0/self.rad.sampling_rate)
    ant_sigs = antennas.antennas_signal(self.ants, self.ant_models, self.sources, timebase)
    rad_sig_full = self.rad.sampled_signal(ant_sigs[0, :], 0)
    obs_full = self.rad.get_full_obs(ant_sigs, self.utc_date, self.config)

    ant_sigs_simp = antennas.antennas_simplified_signal(self.ants, self.ant_models, self.sources, self.rad.baseband_timebase, self.rad.int_freq)
    obs_simp = self.rad.get_simplified_obs(ant_sigs_simp, self.utc_date, self.config)


    freqs, spec_full_before_obs = spectrum.plotSpectrum(rad_sig_full, self.rad.ref_freq, label='full_before_obs_obj', c='blue')
    freqs, spec_full = spectrum.plotSpectrum(obs_full.get_antenna(1), self.rad.ref_freq, label='full', c='cyan')
    freqs, spec_simp = spectrum.plotSpectrum(obs_simp.get_antenna(1), self.rad.ref_freq, label='simp', c='red')
    plt.legend()

    self.assertTrue((spec_full_before_obs == spec_full).all(), True)


    plt.figure()
    plt.plot(freqs, (spec_simp-spec_full)/spec_full)
    plt.show()

    print(len(obs_full.get_antenna(1)), obs_full.get_antenna(1).mean())
    print(len(obs_simp.get_antenna(1)), obs_simp.get_antenna(1).mean())

    # self.assertEqual(self.settings.get_num_antenna(), 5)
    # self.assertLess(d.std(), 0.1)

if __name__ == '__main__':

    a = TestMax2769B()
    a.test_get_obs()




# -*- coding: utf-8 -*-

def telescope_signals(ref_freq, freq_mult, sample_duration, baseline, angle_1, angle_2):
  
  import numpy as np
  
  SNR = 20 # Signal to noise ratio
  
  # Introduce the geometric delay
  delay_1 = baseline*np.sin(angle_1) / 2.99793e8
  delay_2 = baseline*np.sin(angle_2) / 2.99793e8
  
  # how many source rates is this delay
  sample_delay_1 = int(delay_1 * (ref_freq * freq_mult))
  sample_delay_2 = int(delay_2 * (ref_freq * freq_mult))
  
  print('delay_1 = ', delay_1)
  print('delay_2 = ', delay_2)
  print('sample_delay_1 = ', sample_delay_1)
  print('sample_delay_2 = ', sample_delay_2)
  
  # Calculate the number of points required
  n = int(ref_freq * freq_mult * sample_duration)
  
  # create two signals
  source_1 = np.random.random(n + sample_delay_1) # Add gaussian noise
  source_2 = np.random.random(n + sample_delay_2) # Add gaussian noise
  noise_1 = np.random.random(n)/SNR # Add uncorrelated gaussian noise
  noise_2 = np.random.random(n)/SNR # Add uncorrelated gaussian noise
  s1 = source_1[:n] #+ source_2[:n] + noise_1 # Un-delayed signal with instrument noise
  s2 = source_1[sample_delay_1:n+sample_delay_1] #+ source_2[sample_delay_2:n+sample_delay_2]+ noise_2 # Delayed signal with instrument noise
  
  return (s1, s2)

if __name__ == '__main__':
  import numpy as np
  
  ref_freq = 16.368e6
  freq_mult = 256
  sample_duration = 1e-3
  ref_div = 16
  main_div = 1536
  bandwidth = 2.5e6
  order = 5
  baseline = 20.0 # Meters
  angle_1 = 30 * np.pi/180 # Radians
  angle_2 = 60 * np.pi/180 # Radians
  
  telescope_signals(ref_freq, freq_mult, sample_duration, ref_div, main_div, bandwidth, order, baseline, angle_1, angle_2)
